#!/usr/bin/env python
"""
_Execute_

Run the stage out commands in a nice non-blocking way

"""
from __future__ import print_function

from subprocess import Popen, PIPE

from Utils.PythonVersion import PY3
from WMCore.Storage.StageOutError import StageOutError

def runCommand(command):
    """
    _runCommand_

    Run the command without deadlocking stdout and stderr,

    Returns the exitCode

    """
    # capture stdout and stderr from command
    if PY3:
        # python2 pylint complains about `encoding` argument
        child = Popen(command, shell=True, bufsize=1, stdin=PIPE, close_fds=True, encoding='utf8')
    else:
        child = Popen(command, shell=True, bufsize=1, stdin=PIPE, close_fds=True)

    child.communicate()
    retCode = child.returncode

    return retCode

def runCommandWithOutput(command):
    """
    _runCommandWithOutput_

    Run the command without deadlocking stdout and stderr,
    echo all output to sys.stdout and sys.stderr

    Returns the exitCode and the a string containing std out & error

    """
    # capture stdout and stderr from command
    if PY3:
        # python2 pylint complains about `encoding` argument
        child = Popen(command, shell=True, bufsize=1, stdin=PIPE, stdout=PIPE, stderr=PIPE, close_fds=True, encoding='utf8')
    else:
        child = Popen(command, shell=True, bufsize=1, stdin=PIPE, stdout=PIPE, stderr=PIPE, close_fds=True)

    sout, serr = child.communicate()
    retCode = child.returncode
    
    # If child is terminated by signal, err will be negative value. (Unix only)
    sigStr = "Terminated by signal %s\n" % -retCode if retCode < 0 else ""
    output = "%sstdout: %s\nstderr: %s" % (sigStr, sout, serr)
    return retCode, output


def execute(command):
    """
    _execute_

    Execute the command provided, throw a StageOutError if it exits
    non zero

    """
    try:
        exitCode, output = runCommandWithOutput(command)
        msg = "Command exited with status: %s, Output: (%s)" % (exitCode, output)
        print(msg)
    except Exception as ex:
        msg = "Command threw exception: %s" % str(ex)
        print("ERROR: Exception During Stage Out:\n%s" % msg)
        raise StageOutError(msg, Command=command, ExitCode=60311)
    if exitCode:
        msg = "Command exited non-zero: ExitCode:%s \nOutput (%s)" % (exitCode, output)
        print("ERROR: Exception During Stage Out:\n%s" % msg)
        raise StageOutError(msg, Command=command, ExitCode=exitCode)
    return



#!/usr/bin/env python
"""
_GetDeletableWorkflows_

Oracle implementation of Workflow.GetDeletableWorkflows

"""

from WMCore.WMBS.MySQL.Workflow.GetDeletableWorkflows import GetDeletableWorkflows as MySQLGetDeletableWorkflows

class GetDeletableWorkflows(MySQLGetDeletableWorkflows):
    pass



#!/usr/bin/env python
"""
_TestImpl_

Couple of test implementations for unittests

"""

import os
import os.path
import shutil


from WMCore.Storage.StageOutImplV2 import StageOutImplV2
from WMCore.Storage.StageOutError import StageOutFailure




class TestFailImpl(StageOutImplV2):
    """
    _FailImpl_

    Test plugin that always results in a StageOutFailure

    """

    def doTransfer(self, lfn, pfn, stageOut, pnn, command, options, protocol, checksum  ):
        raise StageOutFailure("FailImpl returns FAIL!!!")


    def doDelete(self, lfn, pfn, pnn, command, options, protocol  ):
        raise StageOutFailure("FailImpl returns FAIL!!!")



#!/usr/bin/env python
"""
_SetLocationByLFN_

MySQL implementation of DBSBuffer.SetLocationByLFN
"""

from WMCore.Database.DBFormatter import DBFormatter


class SetLocationByLFN(DBFormatter):
    sql = """INSERT IGNORE INTO dbsbuffer_file_location (filename, location)
               SELECT df.id, dl.id
               FROM dbsbuffer_file df, dbsbuffer_location dl
               WHERE df.lfn = :lfn
               AND dl.pnn = :pnn
    """

    def execute(self, binds, conn=None, transaction=None):
        """
        Expect binds in the form {lfn, pnn}

        """
        self.dbi.processData(self.sql, binds, conn=conn,
                             transaction=transaction)
        return



#!/usr/bin/env python
"""
_ListThresholdsForSubmit_

Query WMBS and ResourceControl to determine how many jobs are still running so
that we can schedule jobs that have just been created.
"""




from WMCore.ResourceControl.MySQL.ThresholdBySite import ThresholdBySite as MySQLThresholdBySite

class ThresholdBySite(MySQLThresholdBySite):
    """
    Oracle version

    """



#!/usr/bin/env python
"""
_GetUnsubscribedDatasets_

Oracle implementation on GetUnsubscribedDatasets.
"""

from __future__ import division

from WMComponent.RucioInjector.Database.MySQL.GetUnsubscribedDatasets import GetUnsubscribedDatasets as MySQLBase


class GetUnsubscribedDatasets(MySQLBase):
    pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
File       : ProcessStats.py
Author     : Valentin Kuznetsov <vkuznet AT gmail dot com>
Description: This module provide ability to get detailed information about
certain UNIX process, referred by PID. It is based on psutil and threading
python modules. It can be used in any web/data-server to get live status
about the running server. Here is an example

.. doctest:

    from Utils.procstats import status threadStack

    class SomeServer(object):
        def __init__(self):
        self.time0 = time.time()
        def status(self):
            sdict = {'server': processStatus()}
            sdict['server'].update({'uptime': time.time()-self.time0})
            sdict.update(threadStack())
            return sdict

The exposed status endpoint will return a json dictionary containing server
upteima and cpu/mem/threads information.
"""

# futures
from __future__ import print_function
from __future__ import division

from future.utils import viewitems

# system modules
import os
import sys
import json
import time
import argparse
import threading
import traceback

try:
    import psutil
except ImportError:
    pass

def _baseProcessStatusFormat(pid=None):
    if not pid:
        pid = os.getpid()
    ttime = time.localtime()
    tstamp = time.strftime('%d/%b/%Y:%H:%M:%S', ttime)
    return {'pid': pid, 'timestamp': tstamp, 'time': time.time()}

def processStatus(pid=None):
    "Return status of the process in a dictionary format"

    pdict = _baseProcessStatusFormat(pid)
    if 'psutil' not in sys.modules:
        return pdict
    proc = psutil.Process(pid)
    pdict.update(proc.as_dict())
    return pdict

def processStatusDict(pid=None):
    "Return status of the process in a dictionary format"
    pdict = _baseProcessStatusFormat(pid)
    if 'psutil' not in sys.modules:
        return pdict
    proc = psutil.Process(pid)
    pdict.update({"cpu_times": dict(proc.cpu_times()._asdict())})
    pdict.update({"cpu_percent": proc.cpu_percent(interval=1.0)})
    pdict.update({"cpu_num": proc.cpu_num()})
    pdict.update({"memory_full_info": dict(proc.memory_full_info()._asdict())})
    pdict.update({"memory_percent": proc.memory_percent()})
    return pdict

def threadStack():
    """
    Return context of all threads in dictionary format where individual
    thread information stored in its own dicts and all threads are groupped
    into threads list. Code based on example from StackOverflow:
    http://stackoverflow.com/questions/132058/showing-the-stack-trace-from-a-running-python-application
    """
    tdict = {}
    id2name = dict([(th.ident, th.name) for th in threading.enumerate()])
    threads = []
    for tid, stack in viewitems(sys._current_frames()):
        tdict = {"thead": id2name.get(tid, ""), "thead_id": tid}
        stacklist = []
        for filename, lineno, name, line in traceback.extract_stack(stack):
            sdict = dict(filename=filename, line_number=lineno, name=name, line=line)
            stacklist.append(sdict)
        tdict.update({"stack": stacklist})
        threads.append(tdict)
    return dict(threads=threads)

def main():
    "Main function to use this module as a stand-along script."
    parser = argparse.ArgumentParser(prog='PROG')
    parser.add_argument("--pid", action="store", dest="pid", help="process id")
    opts = parser.parse_args()

    pdict = processStatus(int(opts.pid))
    pdict.update(threadStack())
    print(json.dumps(pdict))

if __name__ == '__main__':
    main()



#!/usr/bin/env python
"""
_AddIgnore_

MySQL implementation of DBSBufferFiles.AddIgnore
"""

from WMComponent.DBS3Buffer.MySQL.DBSBufferFiles.AddIgnore import AddIgnore as \
    MySQLAddIgnore


class AddIgnore(MySQLAddIgnore):
    sql = """INSERT INTO dbsbuffer_file (lfn, dataset_algo, status, in_phedex)
                SELECT :lfn, :dataset_algo, :status, :in_phedex FROM DUAL WHERE NOT EXISTS
                  (SELECT id FROM dbsbuffer_file WHERE lfn = :lfn)"""



#!/usr/bin/env python
"""
_NewState_

Oracle implementation for labeling a job Complete
"""


from WMCore.BossAir.MySQL.NewState import NewState as MySQLNewState

class NewState(MySQLNewState):
    """
    _NewState_

    Insert new states into bl_status
    """


    sql = """INSERT INTO bl_status (id, name) SELECT bl_status_SEQ.nextval, :name FROM dual
               WHERE NOT EXISTS (SELECT stat.id FROM bl_status stat WHERE stat.name = :name)"""



#!/usr/bin/env python
# pylint: disable=E1101, E1103, C0121
# E1101 doesn't allow you to define config sections using .section_()
# E1103: Transaction attached to myThread


"""
_CreateWorkArea_

Class(es) that create the work area for each jobGroup
Used in JobCreator
"""
from __future__ import division

from builtins import object

import logging
import os
import os.path
import threading
import traceback
from subprocess import Popen, PIPE

from Utils.IteratorTools import grouper
from WMCore.DAOFactory import DAOFactory
from WMCore.WMException import WMException
from WMCore.WMSpec.WMWorkload import WMWorkload, WMWorkloadHelper


def createDirectories(dirList):
    """
    Create the directory if everything is sane

    """
    for sdirList in grouper(dirList, 500):
        cmdArgs = ['mkdir']
        cmdArgs.extend(sdirList)
        pipe = Popen(cmdArgs, stdout=PIPE, stderr=PIPE, shell=False)
        stdout, stderr = pipe.communicate()
        if stderr:
            if hasattr(stderr, "decode"):
                stderr = stderr.decode('utf-8', 'ignore')
            msg = "Error in creating directories: %s\n" % stderr
            logging.error(msg)
            raise CreateWorkAreaException(msg)

    return


def makedirs(directory):
    """
    _makedirs_

    Implementation of os.makedirs
    You can run subscriptions from the same workflow in separate instances
    of the JobCreatorWorker.  This exists to make sure that if they collide,
    a rare but observed possibility, they don't kill everything.
    """

    try:
        os.makedirs(directory)
    except:
        if not os.path.isdir(directory):
            # Then it really screwed up
            msg = "Failed to create directory %s\n" % (directory)
            msg += str(traceback.format_exc())
            logging.error(msg)
            raise CreateWorkAreaException(msg)
        # Else: the directory exists.  Don't complain, but do mention it
        else:
            msg = "Hit error in creating directory %s; ignoring. \n" % (msg)
            msg += str(traceback.format_exc())
            msg = "This looks like an error but everything seems to be in place"
            logging.error(msg)

    return


def getMasterName(startDir, wmWorkload=None, workflow=None):
    """
    Gets a universal name for the jobGroup directory
    Return the uid as the name if none available (THIS SHOULD NEVER HAPPEN)

    """

    if wmWorkload != None:
        workload = wmWorkload.name()
    elif not os.path.exists(workflow.spec):
        msg = "Could not find Workflow spec %s: " % (workflow.spec)
        msg += "Cannot create work area without spec!"
        logging.error(msg)
        raise CreateWorkAreaException(msg)
    else:
        wmWorkload = WMWorkloadHelper(WMWorkload("workload"))
        wmWorkload.load(workflow.spec)

        workload = wmWorkload.name()

    task = workflow.task
    if task.startswith("/" + workload + "/"):
        task = task[len(workload) + 2:]

    return (os.path.join(startDir, workload),
            os.path.join(startDir, workload, task))


class CreateWorkAreaException(WMException):
    """
    This is a totally awesome exception class

    Except, y'know, that it's not.
    """

    pass


class CreateWorkArea(object):
    """
    Basic class for doing the JobMaker dirty work

    """

    def __init__(self, jobGroupID=None, startDir=None):

        myThread = threading.currentThread()
        myThread.logger = logging.getLogger()

        self.jobGroupID = jobGroupID
        self.jobGroup = None
        self.subscript = None
        self.workflow = None
        self.collectionDir = None
        self.wmWorkload = None
        if not startDir:
            self.startDir = os.getcwd()
        else:
            self.startDir = startDir

        self.jobs = {}

        self.getNewJobGroup(jobGroupID)

        return

    def reset(self):
        """
        Reset key variables between runs

        """

        self.workflow = None
        self.wmWorkload = None

    def processJobs(self, jobGroup, startDir=None, wmWorkload=None,
                    workflow=None, transaction=None, conn=None, cache=True):
        """
        Process the work

        This allows you to pass in two pre-loaded objects, the WMWorkloadSpace and the
        WMBS workflow, to save loading time
        """
        self.reset()
        self.wmWorkload = wmWorkload
        self.workflow = workflow
        self.startDir = startDir
        self.transaction = transaction
        self.conn = conn

        self.jobGroup = jobGroup

        # self.getNewJobGroup(jobGroup = jobGroup)
        self.createJobGroupArea()
        self.createWorkArea(cache=cache)

        return

    def getNewJobGroup(self, jobGroup=None):
        """
        This gets a job group passed to the thread
        """

        # See if we actually have a jobGroupID
        if jobGroup:
            self.jobGroupID = jobGroup.id
        else:
            # Then we have no jobGroup
            return


            # We need the workflow to get the spec
            # if self.workflow == None:
            # If we have something in the workflow,
            # assume we were passed a loaded workflow
            # We need the subscription mostly to get the workflow
            # subscript = jobGroup.subscription
            # subscript.load()
            # self.workflow  = subscript['workflow']
            # self.workflow.load()

        # if not jobGroup.exists():
        #    msg = 'JobMaker: Was passed a non-existant Job Group ID %i' % (self.jobGroupID)
        #    logging.error(msg)
        #    raise Exception(msg)

        self.jobGroup = jobGroup

        return

    def createJobGroupArea(self):
        """
        Creates an area for the task which is just the jobGroupUID,
        in which the jobs will be put

        """

        workloadDir, taskDir = getMasterName(startDir=self.startDir,
                                             wmWorkload=self.wmWorkload,
                                             workflow=self.workflow)

        # Create the workload directory
        if not os.path.isdir(workloadDir):
            makedirs(directory=workloadDir)

        # Create the task directory
        if not os.path.isdir(taskDir):
            makedirs(directory=taskDir)

        return

    def createWorkArea(self, cache=True):
        """
        This should handle the master tasks of creating a working area
        It should take a valid jobGroup and call the
        functions that create the components

        """
        myThread = threading.currentThread()

        if self.jobGroup == None:
            msg = 'JobMaker: Ordered to create for non-existant jobGroup  Failing'
            logging.error(msg)
            raise CreateWorkAreaException(msg)

        workloadDir, taskDir = getMasterName(startDir=self.startDir,
                                             wmWorkload=self.wmWorkload,
                                             workflow=self.workflow)
        jobCounter = 0
        nameList = []

        if cache:
            factory = DAOFactory("WMCore.WMBS", myThread.logger, myThread.dbi)
            setBulkCache = factory(classname="Jobs.SetCache")
        nameDictList = []

        # Now actually start to do things
        for job in self.jobGroup.jobs:
            jid = job['id']

            if jobCounter % 1000 == 0:
                # Create a new jobCollection
                # Increment jobCreator if there's already something there
                jobCounter += self.createJobCollection(jobCounter, taskDir)

            jobCounter = jobCounter + 1

            name = self.getDirectoryName(jid)
            nameList.append(name)
            nameDictList.append({'jobid': jid, 'cacheDir': name})
            job['cache_dir'] = name

        if cache:
            setBulkCache.execute(jobDictList=nameDictList,
                                 conn=self.conn,
                                 transaction=self.transaction)

        createDirectories(nameList)

        # change permissions. See #3623
        for directory in nameList:
            os.chmod(directory, 0o775)

        return

    def createJobCollection(self, jobCounter, taskDir):
        """

        Create a sub-directory to allow storage of large jobs
        """

        value = int(jobCounter // 1000)
        jobCollDir = '%s/JobCollection_%i_%i' % (taskDir, self.jobGroup.id, value)
        # Set this to a global variable
        self.collectionDir = jobCollDir
        if not os.path.exists(jobCollDir):
            # This should be the only application
            # You return 0 because the directory you just made should be empty
            os.mkdir(jobCollDir)
            return 0
        if os.path.isdir(jobCollDir):
            # This should never happen
            return len(os.listdir(jobCollDir))
        elif os.path.isfile(jobCollDir):
            # Well, you're screwed.  Some other file is in the way: IN A DIRECTORY YOU JUST CREATED.
            # Time to freak the hell out
            msg = "Could not create jobCollection %s;" % (jobCollDir)
            msg += "non-directory file in the way!"
            logging.error(msg)
            raise CreateWorkAreaException(msg)
        else:
            # You're screwed
            msg = "Something is in the way at %s" % (jobCollDir)
            msg += "File system could not determine type of object"
            logging.error(msg)
            raise CreateWorkAreaException(msg)

    def getDirectoryName(self, jid):
        """
        Gets a universal name for the directory we're working in.

        """

        name = 'job_%i' % (jid)

        return os.path.join(self.collectionDir, name)



"""
_Create_

Base class for creating the BossAir database
"""
from __future__ import print_function

import threading

from WMCore.Database.DBCreator import DBCreator
from WMCore.WMException import WMException


class Create(DBCreator):
    """
    This should create the BossAir schema; since they don't do it.

    """
    sequence_tables = ['bl_runjob', 'bl_status']

    def __init__(self, logger=None, dbi=None, params=None):
        """
        _init_

        Call the DBCreator constructor and create the list of required tables.
        """

        myThread = threading.currentThread()

        if logger is None:
            logger = myThread.logger
        if dbi is None:
            dbi = myThread.dbi

        DBCreator.__init__(self, logger, dbi)

        tablespaceTable = ""
        tablespaceIndex = ""

        self.create = {}
        self.constraints = {}
        self.indexes = {}

        if params:
            if "tablespace_table" in params:
                tablespaceTable = "TABLESPACE %s" % params["tablespace_table"]
            if "tablespace_index" in params:
                tablespaceIndex = "USING INDEX TABLESPACE %s" % params["tablespace_index"]

        self.requiredTables = ["01bl_status", "02bl_runjob"]

        self.sequence_tables = []
        self.sequence_tables.append("bl_status")
        self.sequence_tables.append("bl_runjob")

        self.create['01bl_status'] = \
            """CREATE TABLE bl_status (
                id            INTEGER  NOT NULL,
                name          VARCHAR(255)
                ) %s  """ % (tablespaceTable)

        self.create['02bl_runjob'] = \
            """CREATE TABLE bl_runjob
               (
               id            INTEGER NOT NULL,
               wmbs_id       INTEGER,
               grid_id       VARCHAR(255),
               bulk_id       VARCHAR(255),
               status        CHAR(1)   DEFAULT '1',
               sched_status  INTEGER NOT NULL,
               retry_count   INTEGER,
               status_time   INTEGER,
               location      INTEGER,
               user_id       INTEGER
               ) %s  """ % (tablespaceTable)

        self.indexes["01_pk_bl_status"] = \
            """ALTER TABLE bl_status ADD
                 (CONSTRAINT bl_status_pk PRIMARY KEY (id) %s)""" % tablespaceIndex

        self.indexes["02_pk_bl_runjob"] = \
            """ALTER TABLE bl_runjob ADD
                 (CONSTRAINT bl_runjob_pk PRIMARY KEY (id) %s)""" % tablespaceIndex

        self.constraints["01_fk_bl_runjob"] = \
            """ALTER TABLE bl_runjob ADD
               (CONSTRAINT bl_runjob_fk1 FOREIGN KEY(wmbs_id)
               REFERENCES wmbs_job(id) ON DELETE CASCADE)"""

        self.constraints["01_idx_bl_runjob"] = \
            """CREATE INDEX idx_bl_runjob_wmbs ON bl_runjob(wmbs_id) %s""" % tablespaceIndex

        self.constraints["02_fk_bl_runjob"] = \
            """ALTER TABLE bl_runjob ADD
               (CONSTRAINT bl_runjob_fk2 FOREIGN KEY(sched_status)
               REFERENCES bl_status(id) ON DELETE CASCADE)"""

        self.constraints["02_idx_bl_runjob"] = \
            """CREATE INDEX idx_bl_runjob_status ON bl_runjob(sched_status) %s""" % tablespaceIndex

        self.constraints["03_fk_bl_runjob"] = \
            """ALTER TABLE bl_runjob ADD
               (CONSTRAINT bl_runjob_fk3 FOREIGN KEY(user_id)
               REFERENCES wmbs_users(id) ON DELETE CASCADE)"""

        self.constraints["03_idx_bl_runjob"] = \
            """CREATE INDEX idx_bl_runjob_users ON bl_runjob(user_id) %s""" % tablespaceIndex

        self.constraints["04_fk_bl_runjob"] = \
            """ALTER TABLE bl_runjob ADD
               (CONSTRAINT bl_runjob_fk4 FOREIGN KEY(location)
               REFERENCES wmbs_location(id) ON DELETE CASCADE)"""

        self.constraints["04_idx_bl_runjob"] = \
            """CREATE INDEX idx_bl_runjob_location ON bl_runjob(location) %s""" % tablespaceIndex

        j = 50
        for i in self.sequence_tables:
            seqname = '%s_SEQ' % i
            self.create["%s%s" % (j, seqname)] = \
                "CREATE SEQUENCE %s start with 1 increment by 1 nomaxvalue cache 100" \
                % seqname

        return

    def execute(self, conn=None, transaction=None):
        """
        _execute_

        Check to make sure that all required tables have been defined.  If
        everything is in place have the DBCreator make everything.
        """
        for requiredTable in self.requiredTables:
            if requiredTable not in self.create:
                raise WMException("The table '%s' is not defined." % \
                                  requiredTable, "WMCORE-2")

        try:
            DBCreator.execute(self, conn, transaction)
            return True
        except Exception as e:
            print("ERROR: %s" % e)
            return False



#!/usr/bin/env python
"""
_JobAccountantPoller_

Poll WMBS for complete jobs and process their framework job reports.
"""

import threading
import logging

from Utils.IteratorTools import grouper
from Utils.Timers import timeFunction
from WMCore.WorkerThreads.BaseWorkerThread import BaseWorkerThread
from WMCore.Database.CouchUtils import CouchConnectionError
from WMCore.DAOFactory import DAOFactory
from WMComponent.JobAccountant.AccountantWorker import AccountantWorker
from WMCore.WMException import WMException


class JobAccountantPollerException(WMException):
    """
    _JobAccountantPollerException_

    JobAccountant error class for general
    errors in the poller (used as a catchall
    from the worker).
    """


class JobAccountantPoller(BaseWorkerThread):
    def __init__(self, config):
        BaseWorkerThread.__init__(self)
        self.config = config
        self.accountantWorkSize = getattr(self.config.JobAccountant, 'accountantWorkSize', 100)

        return

    def setup(self, parameters=None):
        """
        _setup_

        Instantiate the requisite number of accountant workers and create a
        processpool with them.  Also instantiate all the DAOs that we will use.
        """
        self.accountantWorker = AccountantWorker(config=self.config)

        myThread = threading.currentThread()
        daoFactory = DAOFactory(package="WMCore.WMBS", logger=myThread.logger,
                                dbinterface=myThread.dbi)
        self.getJobsAction = daoFactory(classname="Jobs.GetFWJRByState")
        return

    @timeFunction
    def algorithm(self, parameters=None):
        """
        _algorithm_

        Poll WMBS for jobs in the 'Complete' state and then pass them to the
        accountant worker.
        """
        completeJobs = self.getJobsAction.execute(state="complete")
        logging.info("Found %d completed jobs", len(completeJobs))

        if len(completeJobs) == 0:
            logging.debug("No work to do; exiting")
            return

        for jobsSlice in grouper(completeJobs, self.accountantWorkSize):
            try:
                self.accountantWorker(jobsSlice)
            except WMException:
                myThread = threading.currentThread()
                if getattr(myThread, 'transaction', None) is not None:
                    myThread.transaction.rollback()
                raise
            except CouchConnectionError as ex:
                msg = "Caught CouchConnectionError exception. Waiting until the next polling cycle.\n"
                msg += str(ex)
                logging.error(msg)
                myThread = threading.currentThread()
                if getattr(myThread, 'transaction', None) is not None:
                    myThread.transaction.rollback()
            except Exception as ex:
                myThread = threading.currentThread()
                if getattr(myThread, 'transaction', None) is not None:
                    myThread.transaction.rollback()
                msg = "Hit general exception in JobAccountantPoller while using worker.\n"
                msg += str(ex)
                logging.exception(msg)
                raise JobAccountantPollerException(msg)

        return



#!/usr/bin/env python
"""
_JobsByState_

Monitoring DAO classes for Jobs in WMBS
"""
__all__ = []



from WMCore.WMBS.MySQL.Monitoring.JobsByState import JobsByState \
    as JobsByStateMySQL

class JobsByState(JobsByStateMySQL):

    #TO check what else is needed for return item
    sql = JobsByStateMySQL.sql



#!/usr/bin/env python
"""
_ListAlgo_

MySQL implementation of DBSBuffer.ListAlgo
"""




from WMCore.Database.DBFormatter import DBFormatter

class ListAlgo(DBFormatter):
    """
    _ListAlgo_

    Retrieve information about an algorithm from the DBSBuffer.  This is mainly
    used by the unit tests to verify that the NewAlgo DAO is working correctly.
    """
    sql = """SELECT id, app_name, app_ver, app_fam, pset_hash, config_content,
                    in_dbs FROM dbsbuffer_algo
               WHERE app_name = :app_name AND app_ver = :app_ver AND
                     app_fam = :app_fam AND pset_hash = :pset_hash"""

    sqlID = """SELECT id, app_name, app_ver, app_fam, pset_hash, config_content,
                      in_dbs FROM dbsbuffer_algo
                 WHERE id = :id"""

    def execute(self, algoID = None, appName = None, appVer = None, appFam = None,
                psetHash = None, configContent = None, conn = None,
                transaction = False):
        """
        _execute_

        Either the algorithm's ID or the appName, appVer, appFam and psetHash
        must be passed in.
        """
        if algoID == None:
            binds = {"app_name": appName, "app_ver": appVer, "app_fam": appFam,
                     "pset_hash": psetHash}
            result = self.dbi.processData(self.sql, binds, conn = conn,
                                          transaction = transaction)
        else:
            binds = {"id": algoID}
            result = self.dbi.processData(self.sqlID, binds, conn = conn,
                                          transaction = transaction)

        return self.formatDict(result)



#!/usr/bin/env python
"""
_Exists_

MySQL implementation of Workflow.Exists

"""
__all__ = []



from WMCore.Database.DBFormatter import DBFormatter

class Exists(DBFormatter):
    sql = """select id from wmbs_workflow
            where spec = :spec and name = :name and task = :task and owner =
           (SELECT id FROM wmbs_users WHERE cert_dn = :owner AND group_name = :group_name AND role_name = :role_name )"""

    def format(self, result):
        result = DBFormatter.format(self, result)

        if len(result) == 0:
            return False
        else:
            return int(result[0][0])

    def getBinds(self, spec=None, owner=None, group_name = '', role_name = '', name = None, task = None):
        return self.dbi.buildbinds(self.dbi.makelist(owner), 'owner',
                                    self.dbi.buildbinds(self.dbi.makelist(spec), 'spec',
                                    self.dbi.buildbinds(self.dbi.makelist(group_name), 'group_name',
                                    self.dbi.buildbinds(self.dbi.makelist(role_name), 'role_name',
                                    self.dbi.buildbinds(self.dbi.makelist(name), 'name',
                                     self.dbi.buildbinds(self.dbi.makelist(task), 'task'))))))

    def execute(self, spec=None, owner=None, group_name = '', role_name = '', \
                     name = None, task = None, conn = None, transaction = False):
        result = self.dbi.processData(self.sql, self.getBinds(spec, owner, group_name, role_name, name, task),
                         conn = conn, transaction = transaction)
        return self.format(result)



#!/usr/bin/env python




from builtins import object
class FetcherInterface(object):
    """
    _FetcherInterface_

    define standard interface for fetcher implementations.

    Init takes the working directory for the sandbox

    Call invokes the fetcher on a WMTask instance

    """

    def __init__(self):
        self.workingDir = None
        self.cacheDir = None

    def setWorkingDirectory(self, workingDir):
        self.workingDir = workingDir

    def workingDirectory(self):
        return self.workingDir

    def setCacheDirectory(self, cacheDir):
        self.cacheDir = cacheDir

    def cacheDirectory(self):
        return self.cacheDir

    def __call__(self, wmTaskHelper):
        """
        _operator(wmTask)_

        Override to pull infomation from the WMTask, retrieve information and
        then insert it into the sandbox

        """
        msg = "FetcherInterface.__call__(wmTask) not overridden by:\n"
        msg += "%s implementation\n" % self.__class__.__name__
        raise NotImplementedError(msg)



#!/usr/bin/env python
"""
_DBS3Buffer.CreateBlocks_

Create new block in dbsbuffer_block
"""

from WMComponent.DBS3Buffer.MySQL.CreateBlocks import CreateBlocks as MySQLCreateBlocks

class CreateBlocks(MySQLCreateBlocks):

    sql = """INSERT INTO dbsbuffer_block
             (id, dataset_id, blockname, location, status, create_time)
             SELECT dbsbuffer_block_seq.nextval,
                    (SELECT id from dbsbuffer_dataset WHERE path = :dataset),
                    :block,
                    (SELECT id FROM dbsbuffer_location WHERE pnn = :location),
                    :status,
                    :time
             FROM DUAL
             """



#!/usr/bin/env python
"""
_ListSubTypes_

Retrieve a list of all subscription types from WMBS.
"""




from WMCore.Database.DBFormatter import DBFormatter

class ListSubTypes(DBFormatter):
    sql = "SELECT name FROM wmbs_sub_types"

    def format(self, result):
        """
        _format_

        Format the query result into a single list of subscription types.
        """
        results = DBFormatter.format(self, result)

        resultList = []
        for result in results:
            resultList.append(result[0])

        return resultList

    def execute(self, conn = None, transaction = False):
        result = self.dbi.processData(self.sql, conn = conn,
                                      transaction = transaction)
        return self.format(result)



#!/usr/bin/env python
"""
_FailFiles_

MySQL implementation of Subscription.FailFiles
"""

from WMCore.Database.DBFormatter import DBFormatter

class FailFiles(DBFormatter):
    sql = """INSERT INTO wmbs_sub_files_failed (subscription, fileid)
               VALUES(:subscription, :fileid)"""

    delAcq = """DELETE FROM wmbs_sub_files_acquired
                WHERE subscription = :subscription AND
                      fileid = :fileid"""

    delAva = """DELETE FROM wmbs_sub_files_available
                WHERE subscription = :subscription AND
                      fileid = :fileid"""

    delCom = """DELETE FROM wmbs_sub_files_complete
                WHERE subscription = :subscription AND
                      fileid = :fileid"""

    def execute(self, subscription = None, file = None, conn = None,
                transaction = False):
        binds = self.getBinds(subscription = subscription, fileid = file)
        self.dbi.processData(self.sql, binds, conn = conn,
                             transaction = transaction)
        self.dbi.processData(self.delAcq, binds, conn = conn,
                             transaction = transaction)
        self.dbi.processData(self.delAva, binds, conn = conn,
                             transaction = transaction)
        self.dbi.processData(self.delCom, binds, conn = conn,
                             transaction = transaction)
        return



#!/usr/bin/env python
"""
_LoadForErrHandler_

MySQL implementation of Jobs.LoadForErrorHandler.
"""

from WMCore.DataStructs.Run import Run
from WMCore.Database.DBFormatter import DBFormatter
from WMCore.WMBS.File import File

from future.utils import listvalues

class LoadForErrorHandler(DBFormatter):
    """
    _LoadForErrorHandler_

    Retrieve meta data for a job given it's ID.  This includes the name,
    job group and last update time. It also works for the AccountantWorker
    to determine ACDC records for skipped files in successful jobs, this mode
    is used when a file selection is provided.
    """
    sql = """SELECT wmbs_job.id, wmbs_job.jobgroup,
               ww.name as workflow, ww.task as task
             FROM wmbs_job
               INNER JOIN wmbs_jobgroup ON wmbs_jobgroup.id = wmbs_job.jobgroup
               INNER JOIN wmbs_subscription ON wmbs_subscription.id = wmbs_jobgroup.subscription
               INNER JOIN wmbs_workflow ww ON ww.id = wmbs_subscription.workflow
             WHERE wmbs_job.id = :jobid"""

    fileSQL = """SELECT wfd.id, wfd.lfn, wfd.filesize AS size, wfd.events, wfd.first_event,
                   wfd.merged, wja.job AS jobid, wpnn.pnn
                 FROM wmbs_file_details wfd
                 INNER JOIN wmbs_job_assoc wja ON wja.fileid = wfd.id
                 LEFT OUTER JOIN wmbs_file_location wfl ON wfd.id = wfl.fileid
                 LEFT OUTER JOIN wmbs_pnns wpnn ON wpnn.id = wfl.pnn
                 WHERE wja.job = :jobid"""

    parentSQL = """SELECT parent.lfn AS lfn, wfp.child AS id
                     FROM wmbs_file_parent wfp
                     INNER JOIN wmbs_file_details parent ON parent.id = wfp.parent
                     WHERE wfp.child = :fileid AND parent.merged = 1"""

    runLumiSQL = """SELECT fileid, run, lumi, num_events FROM wmbs_file_runlumi_map
                     WHERE fileid = :fileid"""


    def getRunLumis(self, fileBinds, fileList,
                    conn=None, transaction=False):
        """
        _getRunLumis_

        Fetch run/lumi/events information for each file and append Run objects
        to the files information.
        """
        if not fileBinds:
            return

        lumiResult = self.dbi.processData(self.runLumiSQL, fileBinds, conn=conn,
                                          transaction=transaction)
        lumiList = self.formatDict(lumiResult)

        lumiDict = {}
        for l in lumiList:
            lumiDict.setdefault(l['fileid'], [])
            lumiDict[l['fileid']].append(l)

        for f in fileList:
            # Add new runs
            f.setdefault('newRuns', [])

            fileRuns = {}
            if f['id'] in lumiDict:
                for l in lumiDict[f['id']]:
                    run = l['run']
                    lumi = l['lumi']
                    numEvents = l['num_events']
                    fileRuns.setdefault(run, [])
                    fileRuns[run].append((lumi, numEvents))

            for r in fileRuns:
                newRun = Run(runNumber=r)
                newRun.lumis = fileRuns[r]
                f['newRuns'].append(newRun)
        return

    def execute(self, jobID, fileSelection=None, conn=None, transaction=False):
        """
        _execute_

        Execute the SQL for the given job ID and then format and return
        the result.
        fileSelection is a dictionary key'ed by the job id and with a list
        of lfns
        """

        if isinstance(jobID, list) and not len(jobID):
            return []
        elif isinstance(jobID, list):
            binds = jobID
        else:
            binds = [{"jobid": jobID}]

        result = self.dbi.processData(self.sql, binds, conn=conn,
                                      transaction=transaction)
        jobList = self.formatDict(result)

        filesResult = self.dbi.processData(self.fileSQL, binds, conn=conn,
                                           transaction=transaction)
        fileList = self.formatDict(filesResult)

        # special case for skipped files
        if fileSelection:
            fileList = [x for x in fileList if x['lfn'] in fileSelection[x['jobid']]]

        noDuplicateFiles = {}
        fileBinds = []
        for x in fileList:
            # Assemble unique list of binds
            if {'fileid': x['id']} not in fileBinds:
                fileBinds.append({'fileid': x['id']})
                noDuplicateFiles[x['id']] = x

        # only upload to not duplicate files to prevent excessive memory
        self.getRunLumis(fileBinds, listvalues(noDuplicateFiles), conn, transaction)

        parentList = []
        if fileBinds:
            parentResult = self.dbi.processData(self.parentSQL, fileBinds, conn=conn,
                                                transaction=transaction)
            parentList = self.formatDict(parentResult)

        filesForJobs = {}
        for f in fileList:
            jobid = f['jobid']
            filesForJobs.setdefault(jobid, {})

            if f['id'] not in filesForJobs[jobid]:
                wmbsFile = File(id=f['id'])
                # need to update with noDuplicateFiles since this one has run lumi information.

                wmbsFile.update(noDuplicateFiles[f["id"]])
                if f['pnn']:  # file might not have a valid location, or be Null
                    wmbsFile['locations'].add(f['pnn'])
                for r in wmbsFile.pop('newRuns'):
                    wmbsFile.addRun(r)
                for entry in parentList:
                    if entry['id'] == f['id']:
                        wmbsFile['parents'].add(entry['lfn'])
                wmbsFile.pop('pnn', None)  # not needed for anything, just remove it
                wmbsFile.pop('newlocations', None)  # not needed for anything, just remove it
                filesForJobs[jobid][f['id']] = wmbsFile
            elif f['pnn']:
                # If the file is there and it has a location, just add it
                filesForJobs[jobid][f['id']]['locations'].add(f['pnn'])

        for j in jobList:
            j.setdefault('input_files', [])
            if j['id'] in filesForJobs:
                j['input_files'] = listvalues(filesForJobs[j['id']])

        return jobList



#!/usr/bin/env python
"""
_GetCouchID_

Oracle implementation of Jobs.GetCouchID
"""




from WMCore.WMBS.MySQL.Jobs.GetCouchID import GetCouchID as MySQLGetCouchID

class GetCouchID(MySQLGetCouchID):
    """
    Identical to MySQL version.
    """
    pass



#!/usr/bin/env python
"""Various helper functions for workqueue"""

import logging
import os
from WMCore.Services.CRIC.CRIC import CRIC


__dbses = {}


def get_dbs(url):
    """Return DBS object for url"""
    try:
        return __dbses[url]
    except KeyError:
        from WMCore.Services.DBS.DBSReader import DBSReader
        __dbses[url] = DBSReader(url)
        return __dbses[url]

__cric = None
__cmsSiteNames = []


def cmsSiteNames():
    """Get all cms sites"""
    global __cmsSiteNames
    if __cmsSiteNames:
        return __cmsSiteNames
    global __cric
    if not __cric:
        __cric = CRIC()

    try:
        __cmsSiteNames = __cric.getAllPSNs()
    except Exception:
        pass
    return __cmsSiteNames


def makeLocationsList(siteWhitelist, siteBlacklist):
    """
    _makeLocationsList_

    Make a location list based on the intersection between a site white list
    and blacklist, if none specified then all sites are listed.
    """
    sites = cmsSiteNames()
    if siteWhitelist:
        # Just get the CMS sites matching the whitelists
        sites = list(set(sites) & set(siteWhitelist))
    if siteBlacklist:
        # Get all CMS sites less the blacklist
        sites = list(set(sites) - set(siteBlacklist))
    return sites


def queueFromConfig(config):
    """Create a queue from the config object"""
    config = queueConfigFromConfigObject(config)
    if config.WorkQueueManager.level == 'GlobalQueue':
        from WMCore.WorkQueue.WorkQueue import globalQueue
        return globalQueue(**config.WorkQueueManager.queueParams)
    elif config.WorkQueueManager.level == 'LocalQueue':
        from WMCore.WorkQueue.WorkQueue import localQueue
        return localQueue(**config.WorkQueueManager.queueParams)
    else:
        from WMCore.WorkQueue.WorkQueue import WorkQueue
        return WorkQueue(**config.WorkQueueManager.queueParams)


def queueConfigFromConfigObject(config):
    """From a config object create a config dict suitable for a queue object"""
    from os import path
    wqManager = config.section_('WorkQueueManager')

    if not hasattr(wqManager, 'componentDir'):
        wqManager.componentDir = path.join(config.General.WorkDir,
                                           'WorkQueueManager')
    if not hasattr(wqManager, 'namespace'):
        wqManager.namespace = 'WMComponent.WorkQueueManager.WorkQueueManager'
    if not hasattr(wqManager, 'logLevel'):
        wqManager.logLevel = 'INFO'
    if not hasattr(wqManager, 'pollInterval'):
        wqManager.pollInterval = 600

    # WorkQueue config
    if not hasattr(wqManager, 'queueParams'):
        wqManager.queueParams = {}
    qConfig = wqManager.queueParams

    # Rucio-related config
    if hasattr(wqManager, 'rucioUrl'):
        qConfig['rucioUrl'] = wqManager.rucioUrl
    if hasattr(wqManager, 'rucioAuthUrl'):
        qConfig['rucioAuthUrl'] = wqManager.rucioAuthUrl

    if hasattr(wqManager, 'couchurl'):
        qConfig['CouchUrl'] = wqManager.couchurl
    if hasattr(wqManager, 'dbname'):
        qConfig['DbName'] = wqManager.dbname
    if hasattr(wqManager, 'inboxDatabase'):
        qConfig['InboxDbName'] = wqManager.inboxDatabase

    # pull some info we need from other areas of the config
    if "BossAirConfig" not in qConfig and hasattr(config, 'BossAir'):
        qConfig["BossAirConfig"] = config
        qConfig['BossAirConfig'].section_("Agent").agentName = config.Agent.agentName
    if "JobDumpConfig" not in qConfig and hasattr(config, 'JobStateMachine'):
        qConfig["JobDumpConfig"] = config
    if "CacheDir" not in qConfig and getattr(config.WorkQueueManager, 'componentDir', None):
        qConfig['CacheDir'] = os.path.join(config.WorkQueueManager.componentDir, 'cache')

    if 'Team' not in qConfig and hasattr(config.Agent, 'teamName'):
        qConfig['Team'] = config.Agent.teamName
    if 'logger' not in qConfig:
        import threading
        myThread = threading.currentThread()
        if not hasattr(myThread, 'logger'):
            loggingLevelName = getattr(wqManager, 'logLevel', 'INFO')
            logging.basicConfig(format='%(asctime)-15s %(levelname)-8s %(module)s: %(message)s',
                                level=getattr(logging, loggingLevelName))
            myThread.logger = logging.getLogger('workqueue')
        qConfig['logger'] = myThread.logger

    # ReqMgr params
    if not hasattr(wqManager, 'reqMgrConfig'):
        wqManager.reqMgrConfig = {}

    return config



#!/usr/bin/env python
"""
_GetCompletedByFileList_

MySQL implementation of Subscription.IsFileCompleted
"""

__all__ = []



from WMCore.Database.DBFormatter import DBFormatter

class GetCompletedByFileList(DBFormatter):

    """
    returns list of file ids which are in complete status by given list of files
    If it returns the same list as input i
    """
    sql = """SELECT fileid FROM wmbs_sub_files_complete
                  WHERE subscription = :subscription AND fileid = :fileid
           """

    def format(self, result):
        out = []
        for r in result:
            if isinstance(r, int):
                # deal with crappy mysql implementation
                out.append(int(r))
            else:
                for f in r.fetchall():
                    out.append(int(f[0]))
        return out

    def getBinds(self, subscription, files):
        binds = []
        for f in files:
            binds.append({'subscription': subscription,  'fileid': f['id']})
        return binds

    def execute(self, subscription, files, conn = None,
                transaction = False):
        results = self.dbi.processData(self.sql, self.getBinds(subscription, files),
                             conn = conn, transaction = transaction)
        return self.format(results)



# -*- coding: utf-8 -*-
"""
_GetRunningSlots_

MySQL implementation of Locations.GetRunningSlots
Created on Mon Jun 18 12:39:14 2012

@author: dballest
"""

from future.utils import  listvalues

from WMCore.Database.DBFormatter import DBFormatter

class GetRunningSlots(DBFormatter):
    sql = """SELECT running_slots FROM  wmbs_location
             WHERE site_name = :location
             """

    def execute(self, siteName, conn = None, transaction = False):
        binds = {"location": siteName}
        result = self.dbi.processData(self.sql, binds, conn = conn,
                                      transaction = transaction)

        return listvalues(result[0].fetchall()[0])[0]



#!/usr/bin/env python
"""
_GetGroupsByJobState_

Oracle implementation of JobGroup.GetGroupsByJobState
"""




from WMCore.WMBS.MySQL.JobGroup.GetGroupsByJobState import GetGroupsByJobState as MySQLGetGroupsByJobState

class GetGroupsByJobState(MySQLGetGroupsByJobState):
    pass



#!/usr/bin/env python
"""
_RandomSeeder_

Simple seed generator for preset seeds

"""

from WMCore.JobSplitting.Generators.GeneratorInterface import GeneratorInterface

class PresetSeeder(GeneratorInterface):
    """
    Seeder that inserts for each initial seed the value
    sent by the dictionary passed through.

    """
    def __init__(self, **options):
        self.options = options


    def __call__(self, wmbsJob):
        baggage = wmbsJob.getBaggage()
        for x in self.options:
            wmbsJob.addBaggageParameter("PresetSeeder.%s" %(x), self.options[x])



#!/usr/bin/env python
"""
_CompleteInput_

MySQL implementation of Jobs.Complete
"""

from WMCore.Database.DBFormatter import DBFormatter

class CompleteInput(DBFormatter):
    """
    _CompleteInput_

    Mark the input files for a job as complete if and only if all the jobs that
    run over the files have been complete successfully.  This will also remove
    entries from the acquired and failed tables.
    """
    fileSelect = """SELECT job_files.subscriptionid, job_files.fileid, job_files.lfn,
                           job_files.jobid,
                           COUNT(wmbs_job_assoc.job) AS total, SUM(wmbs_job.outcome) AS numsuccess FROM
                      (SELECT wmbs_jobgroup.subscription AS subscriptionid,
                              wmbs_job_assoc.job AS jobid,
                              wmbs_job_assoc.fileid AS fileid,
                              wmbs_file_details.lfn AS lfn FROM wmbs_job_assoc
                         INNER JOIN wmbs_job ON
                           wmbs_job_assoc.job = wmbs_job.id
                         INNER JOIN wmbs_jobgroup ON
                           wmbs_job.jobgroup = wmbs_jobgroup.id
                         INNER JOIN wmbs_file_details ON
                           wmbs_job_assoc.fileid = wmbs_file_details.id
                       WHERE wmbs_job.id = :jobid) job_files
                      INNER JOIN wmbs_job_assoc ON
                        job_files.fileid = wmbs_job_assoc.fileid
                      INNER JOIN wmbs_job ON
                        wmbs_job_assoc.job = wmbs_job.id
                      INNER JOIN wmbs_jobgroup ON
                        wmbs_job.jobgroup = wmbs_jobgroup.id
                    WHERE wmbs_jobgroup.subscription = job_files.subscriptionid
                    GROUP BY job_files.subscriptionid, job_files.fileid, job_files.lfn,
                             job_files.jobid
                    HAVING total = numsuccess"""

    acquiredDelete = """DELETE FROM wmbs_sub_files_acquired
                        WHERE subscription = :subid AND fileid = :fileid"""

    failedDelete = """DELETE FROM wmbs_sub_files_failed
                      WHERE subscription = :subid AND fileid = :fileid"""

    sql = """INSERT IGNORE INTO wmbs_sub_files_complete (fileid, subscription)
               VALUES (:fileid, :subid)
               """

    failSql = """INSERT IGNORE INTO wmbs_sub_files_failed (fileid, subscription)
               VALUES (:fileid, :subid)
               """

    def execute(self, id, lfnsToSkip = None, conn = None, transaction = False):
        if isinstance(id, list):
            binds = []
            for singleID in id:
                binds.append({"jobid": singleID})
        else:
            binds = {"jobid": id}

        result = self.dbi.processData(self.fileSelect, binds, conn = conn,
                                      transaction = transaction)
        possibleDeletes = self.formatDict(result)

        completeBinds = []
        failBinds = []
        for possibleDelete in possibleDeletes:
            if lfnsToSkip and possibleDelete["lfn"] in lfnsToSkip.get(possibleDelete["jobid"], []):
                failBinds.append({"fileid": possibleDelete["fileid"],
                                  "subid": possibleDelete["subscriptionid"]})
            else:
                completeBinds.append({"fileid": possibleDelete["fileid"],
                                      "subid": possibleDelete["subscriptionid"]})

        if len(completeBinds) > 0:
            self.dbi.processData(self.acquiredDelete, completeBinds, conn = conn,
                                 transaction = transaction)
            self.dbi.processData(self.failedDelete, completeBinds, conn = conn,
                                 transaction = transaction)
            self.dbi.processData(self.sql, completeBinds, conn = conn,
                                 transaction = transaction)
        if len(failBinds) > 0:
            self.dbi.processData(self.acquiredDelete, failBinds, conn = conn,
                                 transaction = transaction)
            self.dbi.processData(self.failSql, failBinds, conn = conn,
                                 transaction = transaction)
        return



#!/usr/bin/env python
"""
_ListClosable_

MySQL implementation of Fileset.ListClosable
"""

from WMCore.Database.DBFormatter import DBFormatter

class ListClosable(DBFormatter):
    sql = """SELECT fileset FROM
               (SELECT wmbs_fileset.id AS fileset, SUM(wmbs_parent_fileset.open) AS open_parent_filesets,
                       SUM(files_acquired.total_files) AS total_acquired_files,
                       SUM(files_available.total_files) AS total_available_files,
                       SUM(running_jobs.running_count) AS running_jobs FROM wmbs_fileset
                  INNER JOIN wmbs_workflow_output ON
                    wmbs_fileset.id = wmbs_workflow_output.output_fileset
                  INNER JOIN wmbs_subscription wmbs_parent_subscription ON
                    wmbs_workflow_output.workflow_id = wmbs_parent_subscription.workflow
                  INNER JOIN wmbs_workflow ON
                    wmbs_parent_subscription.workflow = wmbs_workflow.id AND
                    ( wmbs_workflow.injected = 1 OR wmbs_workflow.alt_fs_close = 1 )
                  LEFT OUTER JOIN (SELECT subscription, COUNT(DISTINCT fileid) AS total_files
                              FROM wmbs_sub_files_acquired GROUP BY subscription) files_acquired ON
                    wmbs_parent_subscription.id = files_acquired.subscription
                  LEFT OUTER JOIN (SELECT subscription, COUNT(DISTINCT fileid) AS total_files
                              FROM wmbs_sub_files_available GROUP BY subscription) files_available ON
                    wmbs_parent_subscription.id = files_available.subscription
                  LEFT OUTER JOIN (SELECT subscription, COUNT(wmbs_job.id) AS running_count
                              FROM wmbs_jobgroup
                              LEFT OUTER JOIN wmbs_job ON
                                wmbs_jobgroup.id = wmbs_job.jobgroup
                              LEFT OUTER JOIN wmbs_job_state ON
                                wmbs_job.state = wmbs_job_state.id
                              WHERE wmbs_job_state.name != 'cleanout'
                              GROUP BY subscription) running_jobs ON
                    wmbs_parent_subscription.id = running_jobs.subscription
                  INNER JOIN wmbs_fileset wmbs_parent_fileset ON
                    wmbs_parent_subscription.fileset = wmbs_parent_fileset.id
                WHERE wmbs_fileset.open = 1 GROUP BY wmbs_fileset.id) closeable_filesets
             WHERE closeable_filesets.open_parent_filesets = 0 AND
                   COALESCE(closeable_filesets.running_jobs, 0) = 0 AND
                   COALESCE(closeable_filesets.total_acquired_files, 0) = 0 AND
                   COALESCE(closeable_filesets.total_available_files, 0) = 0"""

    def format(self, results):
        """
        _format_

        Take the array of rows that were returned by the query and format that
        into a single list of open fileset names.
        """
        results = DBFormatter.format(self, results)

        filesetIDs = []
        for result in results:
            filesetIDs.append(int(result[0]))

        return filesetIDs

    def execute(self, conn = None, transaction = False):
        result = self.dbi.processData(self.sql, conn = conn,
                                      transaction = transaction)
        return self.format(result)



#!/usr/bin/env python
"""
_GetJobSlotsByCMSName_

Oracle implementation of Locations.GetJobSlotsByCMSName
"""

from WMCore.WMBS.MySQL.Locations.GetJobSlotsByCMSName \
    import GetJobSlotsByCMSName as MySQLGetJobSlotsByCMSName

class GetJobSlotsByCMSName(MySQLGetJobSlotsByCMSName):
    pass



"""
__ListForJobUpdater__

Oracle implementation of ListForJobUpdater

Created on Apr 16, 2013

@author: dballest
"""

from WMCore.WMBS.MySQL.Workflow.ListForJobUpdater import ListForJobUpdater as MySQLListForJobUpdater

class ListForJobUpdater(MySQLListForJobUpdater):
    pass



#!/usr/bin/env python
"""
Oracle implementation of AddPNNs
"""
from __future__ import division

from WMCore.WMBS.MySQL.Locations.AddPNNs import AddPNNs as AddPNNsMySQL


class AddPNNs(AddPNNsMySQL):

    sql = """INSERT /*+ IGNORE_ROW_ON_DUPKEY_INDEX (wmbs_pnns (pnn)) */
               INTO wmbs_pnns (id, pnn) VALUES (wmbs_pnns_SEQ.nextval, :pnn)"""


#!/usr/bin/env python
"""
A base handler from which all handlers inherit from.
Handlers are mapped to messages in the component.
"""

from builtins import str, object

import logging


class BaseHandler(object):
    """
    A base handler from which all handlers inherit from.
    Handlers are mapped to messages in the component.
    """


    def __init__(self, component):
        self.component = component

    def __call__(self, event, payload):
        msg = """
Overload this method to handel event: %s and payload %s
        """ % (str(event), str(payload))
        logging.info(msg)

    def __str__(self):
        """

        return: string

        String representation of the status of this class.
        """
        return str(self.__class__.__name__)



#!/usr/bin/env python
"""
_GetUninjectedFiles_

Retrieve a list of files that have been injected into DBS but not PhEDEx.
Format the output so that it can easily be injected into PhEDEx.

"""
from __future__ import division

from WMCore.Database.DBFormatter import DBFormatter


class GetUninjectedFiles(DBFormatter):
    sql = """SELECT dbsbuffer_file.lfn AS lfn,
                    dbsbuffer_file.filesize AS filesize,
                    dbsbuffer_block.blockname AS blockname,
                    dbsbuffer_dataset.path AS dataset,
                    dbsbuffer_location.pnn AS location,
                    dbsbuffer_file_checksums.cksum as cksum,
                    dbsbuffer_checksum_type.type as cktype
                    FROM dbsbuffer_file
               INNER JOIN dbsbuffer_file_checksums ON
                 dbsbuffer_file.id = dbsbuffer_file_checksums.fileid
               INNER JOIN dbsbuffer_checksum_type ON
                 dbsbuffer_file_checksums.typeid = dbsbuffer_checksum_type.id
               INNER JOIN dbsbuffer_algo_dataset_assoc ON
                 dbsbuffer_file.dataset_algo = dbsbuffer_algo_dataset_assoc.id
               INNER JOIN dbsbuffer_dataset ON
                 dbsbuffer_algo_dataset_assoc.dataset_id = dbsbuffer_dataset.id
               INNER JOIN dbsbuffer_block ON
                 dbsbuffer_file.block_id = dbsbuffer_block.id
               INNER JOIN dbsbuffer_file_location ON
                 dbsbuffer_file.id = dbsbuffer_file_location.filename
               INNER JOIN dbsbuffer_location ON
                 dbsbuffer_file_location.location = dbsbuffer_location.id
             WHERE dbsbuffer_file.in_phedex = 0"""

    def format(self, result):
        """
        _format_

        Format the query results into something that resembles the XML format
        PhEDEx expects for injection:

        {"location1":
          {"dataset1":
            {"block1": {"is-open": "y", "files":
              [{"lfn": "lfn1", "size": 10, "checksum": {"cksum": 4321}},
               {"lfn": "lfn2", "size": 20, "checksum": {"cksum": 4321}]}}}}

        In order to do this, we have to graph the checksum.
        """
        dictResult = DBFormatter.formatDict(self, result)
        self.specCache = {}
        formattedResult = {}
        for row in dictResult:
            location = row['location']

            if location not in formattedResult:
                formattedResult[location] = {}

            locationDict = formattedResult[location]
            if row["dataset"] not in locationDict:
                locationDict[row["dataset"]] = {}

            datasetDict = locationDict[row["dataset"]]
            if row["blockname"] not in datasetDict:
                datasetDict[row["blockname"]] = {"is-open": "y",
                                                 "files": []}

            blockDict = datasetDict[row["blockname"]]
            for fileInfo in blockDict["files"]:
                if fileInfo["lfn"] == row["lfn"]:
                    fileInfo["checksum"][row["cktype"]] = row["cksum"]
                    break
            else:
                cksumDict = {row["cktype"]: row["cksum"]}
                blockDict["files"].append({"lfn": row["lfn"],
                                           "size": row["filesize"],
                                           "checksum": cksumDict})

        return formattedResult

    def execute(self, conn=None, transaction=False):
        result = self.dbi.processData(self.sql, conn=conn,
                                      transaction=transaction)
        return self.format(result)



#!/usr/bin/env python
"""
_GetOutputMap_

Oracle implementation of Jobs.GetOutputMap
"""

from WMCore.WMBS.MySQL.Jobs.GetOutputMap import GetOutputMap as MySQLGetOutputMap

class GetOutputMap(MySQLGetOutputMap):
    pass



#!/usr/bin/env python
"""
Oracle implementation of AddRunLumi
"""

#This has been modified for Oracle

from WMComponent.DBS3Buffer.MySQL.DBSBufferFiles.AddRunLumi import AddRunLumi as MySQLAddRunLumi

class AddRunLumi(MySQLAddRunLumi):

    sql = """INSERT INTO dbsbuffer_file_runlumi_map (filename, run, lumi, num_events)
            select id, :run, :lumi, :num_events from dbsbuffer_file
            where lfn = :lfn"""



#!/usr/bin/env python
"""
_ListDataset_

Oracle implementation of DBSBuffer.ListDataset
"""




from WMComponent.DBS3Buffer.MySQL.ListDataset import ListDataset as MySQLListDataset

class ListDataset(MySQLListDataset):
    """
    _ListDataset_

    Retrieve information about a dataset in the DBSBuffer.  This is mostly used
    by the unit tests.
    """
    pass



#!/usr/bin/env python
"""
_GetCountByState_

Oracle implementation of Jobs.GetCountByState
"""
from __future__ import division

from WMCore.WMBS.MySQL.Jobs.GetCountByState import GetCountByState as MySQLGetCountByState


class GetCountByState(MySQLGetCountByState):
    """
    Identical to MySQL version.
    """
    pass



#!/usr/bin/env python
"""
_AlcaHarvest_

Template for a AlcaHarvest Step

"""

from WMCore.WMSpec.Steps.Template import Template
from WMCore.WMSpec.Steps.Template import CoreHelper
from WMCore.WMSpec.ConfigSectionTree import nodeName


class AlcaHarvestStepHelper(CoreHelper):
    """
    _AlcaHarvestStepHelper_

    Add API calls and helper methods to the basic WMStepHelper to specialise
    for AlcaHarvest tasks

    """
    def setRunNumber(self, runNumber):
        """
        Sets the run number, used to create a
        run specific subdir for the output
        """
        self.data.condition.runNumber = runNumber

    def setConditionOutputLabel(self, outLabel):
        """
        Sets the output label for the fake
        output model containing the sqlite files
        """
        self.data.condition.outLabel = outLabel

    def setConditionLFNBase(self, lfnbase):
        """
        Sets the LFNBase for condition file copy
        (within the EOSCMS CERN EOS instance)
        """
        self.data.condition.lfnbase = lfnbase

    def setLuminosityURL(self, url):
        """
        Sets the ROOT URL for luminosity file copy
        """
        self.data.luminosity.url = url

class AlcaHarvest(Template):
    """
    _AlcaHarvest_

    Tools for creating a template AlcaHarvest Step

    """

    def install(self, step):
        stepname = nodeName(step)
        step.stepType = "AlcaHarvest"
        step.section_("condition")
        step.condition.runNumber = None
        step.condition.outLabel = None
        step.condition.lfnbase = None
        step.section_("luminosity")
        step.luminosity.url = None

    def helper(self, step):
        """
        _helper_

        Wrap the WMStep provided in the CMSSW helper class that
        includes the ability to add and manipulate the details
        of a CMSSW workflow step

        """
        return AlcaHarvestStepHelper(step)



#!/usr/bin/env python
"""
_GetCompletedWorkflows_

"""

from WMCore.Database.DBFormatter import DBFormatter

class GetCompletedWorkflows(DBFormatter):
    """
    _GetCompletedWorkflows_
    """
    sql = """SELECT name, count(name), sum(completed)
                FROM dbsbuffer_workflow
                GROUP BY name"""

    def execute(self, conn=None, transaction=False):
        """
        """
        result = self.dbi.processData(self.sql, conn = conn,
                                          transaction = transaction)
        workflowList = []
        for row in self.format(result):
            if row[1] == row[2]:
                workflowList.append(row[0])
        return workflowList



#!/usr/bin/env python
"""
_ListByState_

MySQL implementation of ListByState

Created on July 10, 2013

@author: tsarangi
"""

from WMCore.Database.DBFormatter import DBFormatter

class ListByState(DBFormatter):
    """
    DAO to list jobs given certain state only,
    """
    sql = """SELECT wmbs_job.id, wmbs_job.retry_count,
                    wmbs_job.cache_dir
                FROM wmbs_job
                INNER JOIN wmbs_job_state ON
                    wmbs_job.state = wmbs_job_state.id
                 WHERE wmbs_job_state.name = :state"""

    def execute(self, state, conn = None, transaction = False):
        result = self.dbi.processData(self.sql, {'state' : state},
                                      conn = conn,
                                      transaction = transaction)
        return self.formatDict(result)



#!/usr/bin/env python
"""
_StepChain_

Creates a StepChain like workflow where there is only one production/processing
task and an arbitrary number of steps (cmsRun) inside this task, which are
chained based on their StepName, InputStep and InputFromOutputModule.

Each Step is formed as a dictionary providing only the basic information, see:
test/data/ReqMgr/requests/DMWM/StepChain_MC.json

It also assumes all the intermediate steps output are transient and do not need
to be staged out and registered in DBS/PhEDEx. Only the last step output will be
made available.
"""
from __future__ import division
from future.utils import viewitems
from builtins import range

from Utils.Utilities import strToBool
import WMCore.WMSpec.Steps.StepFactory as StepFactory
from WMCore.Lexicon import primdataset, taskStepName
from WMCore.WMSpec.StdSpecs.StdBase import StdBase
from WMCore.WMSpec.WMWorkloadTools import (validateArgumentsCreate, parsePileupConfig,
                                           checkMemCore, checkEventStreams, checkTimePerEvent)
from WMCore.WMSpec.WMSpecErrors import WMSpecFactoryException


class StepChainWorkloadFactory(StdBase):
    """
    __StepChainWorkloadFactory__

    Factory for StepChain workflows.
    """

    def __init__(self):
        """
        __init__

        Setup parameters that will be later overwritten in the call,
        otherwise pylint will complain about them.
        """
        StdBase.__init__(self)
        self.configCacheUrl = None
        self.globalTag = None
        self.frameworkVersion = None
        self.scramArch = None
        self.couchDBName = None
        self.stepChain = None
        self.sizePerEvent = None
        self.timePerEvent = None
        self.primaryDataset = None
        self.prepID = None
        self.eventsPerJob = None
        self.eventsPerLumi = None
        # stepMapping is going to be used during assignment for properly mapping
        # the arguments to each step/cmsRun
        self.stepMapping = {}
        self.stepParentageMapping = {}

    def __call__(self, workloadName, arguments):
        """
        __call__

        Create a StepChain workload with the given parameters.
        Configures the workload based on the first task information,
        then properly sets up the remaining tasks.
        """
        StdBase.__call__(self, workloadName, arguments)
        self.workload = self.createWorkload()

        # Update the task configuration
        taskConf = {}
        for k, v in viewitems(arguments["Step1"]):
            taskConf[k] = v
        self.modifyTaskConfiguration(taskConf, True, 'InputDataset' not in taskConf)

        self.inputPrimaryDataset = self.getStepValue('PrimaryDataset', taskConf, self.primaryDataset)
        self.blockBlacklist = taskConf["BlockBlacklist"]
        self.blockWhitelist = taskConf["BlockWhitelist"]
        self.runBlacklist = taskConf["RunBlacklist"]
        self.runWhitelist = taskConf["RunWhitelist"]
        self.splittingAlgo = taskConf['SplittingAlgo']

        # Create the first task
        firstTask = self.workload.newTask(taskConf['StepName'])

        # it has to be called before the other steps are created
        self.createStepMappings(arguments)

        # Create a proper task and set workload level arguments
        startPolicy = self.decideWorkQueueStartPolicy(arguments)
        self.workload.setWorkQueueSplitPolicy(startPolicy, taskConf['SplittingAlgo'],
                                              taskConf['SplittingArguments'],
                                              OpenRunningTimeout=self.openRunningTimeout)
        if startPolicy == "MonteCarlo":
            self.workload.setDashboardActivity("production")
            self.workload.setEndPolicy("SingleShot")
            self.setupGeneratorTask(firstTask, taskConf)
        else:
            self.workload.setDashboardActivity("processing")
            self.setupTask(firstTask, taskConf)

        # Now modify this task to add the next steps
        if self.stepChain > 1:
            self.setupNextSteps(firstTask, arguments)

        self.createStepParentageMappings(firstTask, arguments)

        self.workload.setStepMapping(self.stepMapping)
        self.workload.setStepParentageMapping(self.stepParentageMapping)
        # and push the parentage map to the reqmgr2 workload cache doc
        arguments['ChainParentageMap'] = self.workload.getChainParentageSimpleMapping()

        # Feed values back to save in couch
        if self.eventsPerJob:
            arguments['Step1']['EventsPerJob'] = self.eventsPerJob
        if self.eventsPerLumi:
            arguments['Step1']['EventsPerLumi'] = self.eventsPerLumi
        return self.workload

    def createStepMappings(self, origArgs):
        """
        _createStepMappings_

        Create a simple map of StepName to Step and cmsRun number.
        cmsRun numbers are sequential, just like the step number.
        :param origArgs: arguments provided by the user + default spec args
        :return: update a dictionary in place which is latter used to set a
         `stepMapping` property in the workload object
        """
        for i in range(1, self.stepChain + 1):
            stepNumber = "Step%d" % i
            stepName = origArgs[stepNumber]['StepName']
            cmsRunNumber = "cmsRun%d" % i
            self.stepMapping.setdefault(stepName, (stepNumber, cmsRunNumber))

    def createStepParentageMappings(self, firstTaskO, origArgs):
        """
        _createStepParentageMappings_

        Create a dict struct with a mapping of step name to parent step. It
        also includes a map of output datasets and parent dataset.
        :param firstTaskO: a WMTask object with the top level StepChain task
        :param origArgs: arguments provided by the user + default spec args
        :return: update a dictionary in place which will be later set as a
        WMWorkload property
        """
        for i in range(1, self.stepChain + 1):
            stepNumber = "Step%d" % i
            stepName = origArgs[stepNumber]['StepName']
            cmsRunNumber = "cmsRun%d" % i

            self.stepParentageMapping.setdefault(stepName, {})
            self.stepParentageMapping[stepName] = {'StepNumber': stepNumber,
                                                   'StepCmsRun': cmsRunNumber,
                                                   'ParentStepName': None,
                                                   'ParentStepNumber': None,
                                                   'ParentStepCmsRun': None,
                                                   'ParentDataset': None,
                                                   'OutputDatasetMap': {}}

            if stepNumber == 'Step1':
                self.stepParentageMapping[stepName]['ParentDataset'] = origArgs[stepNumber].get('InputDataset')

            # set the OutputDatasetMap or empty if KeepOutput is False
            if origArgs[stepNumber].get("KeepOutput", True):
                stepHelper = firstTaskO.getStepHelper(cmsRunNumber)
                for outputModuleName in stepHelper.listOutputModules():
                    outputModule = stepHelper.getOutputModule(outputModuleName)
                    outputDataset = "/%s/%s/%s" % (outputModule.primaryDataset,
                                                   outputModule.processedDataset,
                                                   outputModule.dataTier)
                    self.stepParentageMapping[stepName]['OutputDatasetMap'][outputModuleName] = outputDataset

            if "InputStep" in origArgs[stepNumber]:
                parentStepName = origArgs[stepNumber]["InputStep"]
                self.stepParentageMapping[stepName]['ParentStepName'] = parentStepName
                parentStepNumber = self.stepParentageMapping[parentStepName]['StepNumber']
                self.stepParentageMapping[stepName]['ParentStepNumber'] = parentStepNumber
                parentStepCmsRun = self.stepParentageMapping[parentStepName]['StepCmsRun']
                self.stepParentageMapping[stepName]['ParentStepCmsRun'] = parentStepCmsRun

                parentOutputModName = origArgs[stepNumber]["InputFromOutputModule"]
                parentDset = self.findParentStepWithOutputDataset(origArgs, parentStepNumber, parentStepName, parentOutputModName)
                self.stepParentageMapping[stepName]['ParentDataset'] = parentDset

    def findParentStepWithOutputDataset(self, origArgs, stepNumber, stepName, outModName):
        """
        _findParentStepWithOutputDataset_
        Given the parent step name and output module name, finds the parent dataset 
        :param origArgs: request arguments
        :param stepNumber: step number of the parent step
        :param stepName: step name of the parent step
        :param outModName: output module name of the parent step
        :return: the parent dataset name (str), otherwise None
        """
        if origArgs[stepNumber].get("KeepOutput", True):
            return self.stepParentageMapping[stepName]['OutputDatasetMap'][outModName]
        else:
            # then fetch grand-parent data
            parentStepNumber = self.stepParentageMapping[stepName]['ParentStepNumber']
            parentStepName = self.stepParentageMapping[stepName]['ParentStepName']
            if parentStepNumber:
                parentOutputModName = origArgs[stepNumber]["InputFromOutputModule"]
                return self.findParentStepWithOutputDataset(origArgs, parentStepNumber, parentStepName, parentOutputModName)
            else:
                # this is Step1, return the InputDataset if any
                return origArgs[stepNumber].get("InputDataset")

    def setupGeneratorTask(self, task, taskConf):
        """
        _setupGeneratorTask_

        Set up an initial generator task for the top level step (cmsRun1)
        """
        configCacheID = taskConf['ConfigCacheID']
        splitAlgorithm = taskConf["SplittingAlgo"]
        splitArguments = taskConf["SplittingArguments"]
        outMods = self.setupProcessingTask(task, "Production", couchDBName=self.couchDBName,
                                           configDoc=configCacheID, configCacheUrl=self.configCacheUrl,
                                           splitAlgo=splitAlgorithm, splitArgs=splitArguments,
                                           seeding=taskConf['Seeding'],
                                           totalEvents=taskConf['RequestNumEvents'],
                                           cmsswVersion=taskConf.get("CMSSWVersion", None),
                                           scramArch=taskConf.get("ScramArch", None),
                                           globalTag=taskConf.get("GlobalTag", None),
                                           taskConf=taskConf)

        if taskConf["PileupConfig"]:
            self.setupPileup(task, taskConf['PileupConfig'], stepName="cmsRun1")

        # outputModules were added already, we just want to create merge tasks here
        if strToBool(taskConf.get('KeepOutput', True)):
            self.setupMergeTask(task, taskConf, "cmsRun1", outMods)

        return

    def setupTask(self, task, taskConf):
        """
        _setupTask_

        Build the task using the setupProcessingTask from StdBase
        and set the parents appropriately to handle a processing task,
        It's only called for the top level task and top level step (cmsRun1)
        """
        configCacheID = taskConf["ConfigCacheID"]
        splitAlgorithm = taskConf["SplittingAlgo"]
        splitArguments = taskConf["SplittingArguments"]
        self.inputDataset = taskConf["InputDataset"]
        # Use PD from the inputDataset if not provided in the task itself
        if not self.inputPrimaryDataset:
            self.inputPrimaryDataset = self.inputDataset[1:].split("/")[0]

        outMods = self.setupProcessingTask(task, "Processing",
                                           inputDataset=self.inputDataset, couchDBName=self.couchDBName,
                                           configDoc=configCacheID, configCacheUrl=self.configCacheUrl,
                                           splitAlgo=splitAlgorithm, splitArgs=splitArguments,
                                           cmsswVersion=taskConf.get("CMSSWVersion", None),
                                           scramArch=taskConf.get("ScramArch", None),
                                           globalTag=taskConf.get("GlobalTag", None),
                                           taskConf=taskConf)

        lumiMask = taskConf.get("LumiList", self.workload.getLumiList())
        if lumiMask:
            task.setLumiMask(lumiMask)

        if taskConf["PileupConfig"]:
            self.setupPileup(task, taskConf['PileupConfig'], stepName="cmsRun1")

        # outputModules were added already, we just want to create merge tasks here
        if strToBool(taskConf.get('KeepOutput', True)):
            self.setupMergeTask(task, taskConf, "cmsRun1", outMods)

        return

    def setupNextSteps(self, task, origArgs):
        """
        _setupNextSteps_

        Modify the step one task to include N more CMSSW steps and
        chain the output between all three steps.
        """
        self.stepParentageMapping.setdefault(origArgs['Step1']['StepName'], {})

        for i in range(2, self.stepChain + 1):
            currentStepNumber = "Step%d" % i
            currentCmsRun = "cmsRun%d" % i
            taskConf = {}
            for k, v in viewitems(origArgs[currentStepNumber]):
                taskConf[k] = v

            parentStepNumber = self.stepMapping.get(taskConf['InputStep'])[0]
            parentCmsRun = self.stepMapping.get(taskConf['InputStep'])[1]
            parentCmsswStep = task.getStep(parentCmsRun)
            parentCmsswStepHelper = parentCmsswStep.getTypeHelper()

            # Set default values for the task parameters
            self.modifyTaskConfiguration(taskConf, False, 'InputDataset' not in taskConf)
            globalTag = self.getStepValue('GlobalTag', taskConf, self.globalTag)
            frameworkVersion = self.getStepValue('CMSSWVersion', taskConf, self.frameworkVersion)
            scramArch = self.getStepValue('ScramArch', taskConf, self.scramArch)
            prepId = self.getStepValue('PrepID', taskConf, self.prepID)

            currentCmssw = parentCmsswStep.addTopStep(currentCmsRun)
            currentCmssw.setStepType("CMSSW")
            template = StepFactory.getStepTemplate("CMSSW")
            template(currentCmssw.data)

            currentCmsswStepHelper = currentCmssw.getTypeHelper()
            currentCmsswStepHelper.setPrepId(prepId)
            currentCmsswStepHelper.setGlobalTag(globalTag)
            currentCmsswStepHelper.setupChainedProcessing(parentCmsRun, taskConf['InputFromOutputModule'])
            currentCmsswStepHelper.cmsswSetup(frameworkVersion, softwareEnvironment="", scramArch=scramArch)
            currentCmsswStepHelper.setConfigCache(self.configCacheUrl, taskConf['ConfigCacheID'], self.couchDBName)

            # multicore settings
            multicore = self.multicore
            eventStreams = self.eventStreams
            if taskConf['Multicore'] > 0:
                multicore = taskConf['Multicore']
            if taskConf.get("EventStreams") is not None and taskConf['EventStreams'] >= 0:
                eventStreams = taskConf['EventStreams']

            currentCmsswStepHelper.setNumberOfCores(multicore, eventStreams)

            # Pileup check
            taskConf["PileupConfig"] = parsePileupConfig(taskConf["MCPileup"], taskConf["DataPileup"])
            if taskConf["PileupConfig"]:
                self.setupPileup(task, taskConf['PileupConfig'], stepName=currentCmsRun)

            # Handling the output modules in order to decide whether we should
            # stage them out and report them in the Report.pkl file
            parentKeepOutput = strToBool(origArgs[parentStepNumber].get('KeepOutput', True))
            parentCmsswStepHelper.keepOutput(parentKeepOutput)
            childKeepOutput = strToBool(taskConf.get('KeepOutput', True))
            currentCmsswStepHelper.keepOutput(childKeepOutput)
            self.setupOutputModules(task, taskConf, currentCmsRun, childKeepOutput)

        # Closing out the task configuration. The last step output must be saved/merged
        currentCmsswStepHelper.keepOutput(True)

        return

    def getStepValue(self, keyName, stepDict, topLevelValue):
        """
        Utilitarian method to reliably get the value of a step key
        or fallback to the top level one.
        """
        if keyName in stepDict and stepDict.get(keyName) is not None:
            return stepDict.get(keyName)
        else:
            return topLevelValue

    def setupOutputModules(self, task, taskConf, stepCmsRun, keepOutput):
        """
        _setupOutputModules_

        Retrieves the outputModules from the step configuration and sets up
        a merge task for them. Only when KeepOutput is set to True.
        """
        taskConf = taskConf or {}

        outputMods = {}

        configOutput = self.determineOutputModules(configDoc=taskConf["ConfigCacheID"],
                                                   configCacheUrl=self.configCacheUrl,
                                                   couchDBName=self.couchDBName)
        for outputModuleName in configOutput:
            outputModule = self.addOutputModule(task, outputModuleName,
                                                self.inputPrimaryDataset,
                                                configOutput[outputModuleName]["dataTier"],
                                                configOutput[outputModuleName]["filterName"],
                                                stepName=stepCmsRun, taskConf=taskConf)
            outputMods[outputModuleName] = outputModule

        if keepOutput:
            self.setupMergeTask(task, taskConf, stepCmsRun, outputMods)

        return

    def setupMergeTask(self, task, taskConf, stepCmsRun, outputMods):
        """
        _setupMergeTask_

        Adds a merge task to the parent with the proper task configuration.
        """
        frameworkVersion = taskConf.get("CMSSWVersion", self.frameworkVersion)
        scramArch = taskConf.get("ScramArch", self.scramArch)
        # PrepID has to be inherited from the workload level, not from task
        if not taskConf.get('PrepID'):
            taskConf['PrepID'] = self.prepID

        for outputModuleName in outputMods:
            dummyTask = self.addMergeTask(task, self.splittingAlgo, outputModuleName, stepCmsRun,
                                          cmsswVersion=frameworkVersion, scramArch=scramArch,
                                          forceTaskName=taskConf.get('StepName'), taskConf=taskConf)

        return

    def modifyTaskConfiguration(self, taskConf, firstTask=False, generator=False):
        """
        _modifyTaskConfiguration_

        Modify the TaskConfiguration according to the specifications
        in getWorkloadCreateArgs and getChainCreateArgs. It does type
        casting and assigns default values.
        """
        baseArguments = self.getWorkloadCreateArgs()
        for argument in baseArguments:
            if argument in taskConf:
                taskConf[argument] = baseArguments[argument]["type"](taskConf[argument])

        taskArguments = self.getChainCreateArgs(firstTask, generator)
        for argument in taskArguments:
            if argument not in taskConf:
                taskConf[argument] = taskArguments[argument]["default"]
            else:
                taskConf[argument] = taskArguments[argument]["type"](taskConf[argument])

        taskConf["PileupConfig"] = parsePileupConfig(taskConf["MCPileup"], taskConf["DataPileup"])

        if firstTask:
            self.modifyJobSplitting(taskConf, generator)
        return

    def modifyJobSplitting(self, taskConf, generator):
        """
        _modifyJobSplitting_

        Adapt job splitting according to the first step configuration
        or lack of some of them.
        """
        if generator:
            requestNumEvts = int(taskConf.get("RequestNumEvents", 0))
            filterEff = taskConf.get("FilterEfficiency")
            # Adjust totalEvents according to the filter efficiency
            taskConf["SplittingAlgo"] = "EventBased"
            taskConf["RequestNumEvents"] = int(requestNumEvts / filterEff)
            taskConf["SizePerEvent"] = self.sizePerEvent * filterEff

        taskConf["SplittingArguments"] = {}
        if taskConf["SplittingAlgo"] in ["EventBased", "EventAwareLumiBased"]:
            taskConf["EventsPerJob"], taskConf["EventsPerLumi"] = StdBase.calcEvtsPerJobLumi(taskConf.get("EventsPerJob"),
                                                                                             taskConf.get("EventsPerLumi"),
                                                                                             self.timePerEvent,
                                                                                             taskConf.get("RequestNumEvents"))
            self.eventsPerJob = taskConf["EventsPerJob"]
            self.eventsPerLumi = taskConf["EventsPerLumi"]
            taskConf["SplittingArguments"]["events_per_job"] = taskConf["EventsPerJob"]
            if taskConf["SplittingAlgo"] == "EventBased":
                taskConf["SplittingArguments"]["events_per_lumi"] = taskConf["EventsPerLumi"]
            else:
                taskConf["SplittingArguments"]["job_time_limit"] = 48 * 3600  # 2 days
            taskConf["SplittingArguments"]["lheInputFiles"] = taskConf["LheInputFiles"]
        elif taskConf["SplittingAlgo"] == "LumiBased":
            taskConf["SplittingArguments"]["lumis_per_job"] = taskConf["LumisPerJob"]
        elif taskConf["SplittingAlgo"] == "FileBased":
            taskConf["SplittingArguments"]["files_per_job"] = taskConf["FilesPerJob"]

        taskConf["SplittingArguments"].setdefault("include_parents", taskConf['IncludeParents'])
        taskConf["SplittingArguments"].setdefault("deterministicPileup", self.deterministicPileup)

        return

    @staticmethod
    def getWorkloadCreateArgs():
        baseArgs = StdBase.getWorkloadCreateArgs()
        specArgs = {"RequestType": {"default": "StepChain", "optional": False},
                    "Step1": {"default": {}, "optional": False, "type": dict},
                    # ConfigCacheID is not used in the main dict for StepChain
                    "ConfigCacheID": {"optional": True, "null": True},
                    "DeterministicPileup": {"default": False, "type": strToBool, "optional": True, "null": False},
                    "PrimaryDataset": {"null": True, "validate": primdataset},
                    "StepChain": {"default": 1, "type": int, "null": False,
                                  "optional": False, "validate": lambda x: x > 0},
                    "ChainParentageMap": {"default": {}, "type": dict},
                    "FirstEvent": {"default": 1, "type": int, "validate": lambda x: x > 0,
                                   "null": False},
                    "FirstLumi": {"default": 1, "type": int, "validate": lambda x: x > 0,
                                  "null": False},
                    "ParentageResolved": {"default": False, "type": strToBool, "null": False},
                    ### Override StdBase parameter definition
                    "TimePerEvent": {"default": 12.0, "type": float, "validate": lambda x: x > 0},
                    "Memory": {"default": 2300.0, "type": float, "validate": lambda x: x > 0},
                    "Multicore": {"default": 1, "type": int, "validate": checkMemCore},
                    "EventStreams": {"type": int, "null": True, "default": 0, "validate": checkEventStreams}
                   }
        baseArgs.update(specArgs)
        StdBase.setDefaultArgumentsProperty(baseArgs)
        return baseArgs

    @staticmethod
    def getChainCreateArgs(firstTask=False, generator=False):
        """
        _getChainCreateArgs_

        This represents the authoritative list of request arguments that are
        allowed in each chain (Step/Task) of chained request, during request creation.
        Additional especific arguments must be defined inside each spec class.

        For more information on how these arguments are built, please have a look
        at the docstring for getWorkloadCreateArgs.
        """
        baseArgs = StdBase.getChainCreateArgs(firstTask, generator)
        arguments = {
            'InputStep': {'default': None, 'null': False, 'optional': firstTask},
            'StepName': {'null': False, 'optional': False, 'validate': taskStepName},
            'PrimaryDataset': {'default': None, 'optional': True,
                               'validate': primdataset, 'null': False}
            }

        baseArgs.update(arguments)
        StdBase.setDefaultArgumentsProperty(baseArgs)

        return baseArgs

    @staticmethod
    def getWorkloadAssignArgs():
        baseArgs = StdBase.getWorkloadAssignArgs()
        specArgs = {
            "ChainParentageMap": {"default": {}, "type": dict},
            ### Override StdBase assignment parameter definition
            "Memory": {"type": float, "validate": checkMemCore},
            "Multicore": {"type": int, "validate": checkMemCore},
            "EventStreams": {"type": int, "validate": checkEventStreams},
        }
        baseArgs.update(specArgs)
        StdBase.setDefaultArgumentsProperty(baseArgs)
        return baseArgs

    def validateSchema(self, schema):
        """
        _validateSchema_

        Settings that are not supported and will cause workflow injection to fail, are:
         * workflow with more than 10 steps
         * output from the last step *must* be saved
         * each step configuration must be a dictionary
         * StepChain argument must reflect the number of Steps in the request
         * trident configuration, where 2 steps have the same output module AND datatier
         * usual ConfigCacheID validation
         * and the usual Step arguments validation, as defined in the spec
        """
        numSteps = schema['StepChain']
        if numSteps > 10:
            msg = "Workflow exceeds the maximum allowed number of steps. "
            msg += "Limited to up to 10 steps, found %s steps." % numSteps
            self.raiseValidationException(msg)

        lastStep = "Step%s" % schema['StepChain']
        if not strToBool(schema[lastStep].get('KeepOutput', True)):
            msg = "Dropping the output (KeepOutput=False) of the last step is prohibited.\n"
            msg += "You probably want to remove that step completely and try again."
            self.raiseValidationException(msg=msg)

        outputModTier = []
        for i in range(1, numSteps + 1):
            stepNumber = "Step%s" % i
            if stepNumber not in schema:
                msg = "Step '%s' not present in the request schema." % stepNumber
                self.raiseValidationException(msg=msg)

            step = schema[stepNumber]
            # We can't handle non-dictionary steps
            if not isinstance(step, dict):
                msg = "Step '%s' not defined as a dictionary. " % stepNumber
                msg += "It could be an indicator of JSON error.\n"
                self.raiseValidationException(msg=msg)

            # Generic step parameter validation
            self.validateStep(step, self.getChainCreateArgs(i == 1, i == 1 and 'InputDataset' not in step))

            # Validate the existence of the configCache
            if step["ConfigCacheID"]:
                self.validateConfigCacheExists(configID=step['ConfigCacheID'],
                                               configCacheUrl=schema['ConfigCacheUrl'],
                                               couchDBName=schema["CouchDBName"],
                                               getOutputModules=False)
            # we cannot save output of two steps using the same output module and datatier(s)
            if strToBool(step.get("KeepOutput", True)):
                configOutput = self.determineOutputModules(configDoc=step["ConfigCacheID"],
                                                           configCacheUrl=schema['ConfigCacheUrl'],
                                                           couchDBName=schema["CouchDBName"])
                for modName, values in viewitems(configOutput):
                    thisOutput = (modName, values['dataTier'])
                    if thisOutput in outputModTier:
                        msg = "StepChain cannot save output of different steps using "
                        msg += "the same output module AND datatier(s)."
                        msg += "\n%s re-using outputModule: %s and datatier: %s" % (stepNumber,
                                                                                    modName,
                                                                                    values['dataTier'])
                        self.raiseValidationException(msg=msg)
                    outputModTier.append(thisOutput)
        return

    def validateStep(self, taskConf, taskArgumentDefinition):
        """
        _validateStep_

        Validate the step information against the given
        argument description
        """
        try:
            validateArgumentsCreate(taskConf, taskArgumentDefinition, checkInputDset=False)
        except WMSpecFactoryException:
            # just re-raise it to keep the error message clear
            raise
        except Exception as ex:
            self.raiseValidationException(str(ex))

        return

    def decideWorkQueueStartPolicy(self, reqDict):
        """
        Given a request dictionary, decides which WorkQueue start
        policy needs to be used in a given request.
        :param reqDict: a dictionary with the creation request information
        :return: a string with the start policy to be used.
        """
        if not reqDict["Step1"].get("InputDataset"):
            return "MonteCarlo"

        inputDset = reqDict["Step1"]["InputDataset"]
        if inputDset.endswith("/MINIAODSIM"):
            return "Dataset"
        else:
            return "Block"



#!/usr/bin/env python
"""
_StepSpace_

Frontend module for setting up StepSpace areas within a job.

"""
from builtins import object
import os
import inspect
from WMCore.WMRuntime.Sandbox import Sandbox

class StepSpace(object):
    """
    _StepSpace_

    Working area utils for a Step during execution
    TaskSpace reference is set when StepSpace is retrieved from top level
    task space instance

    """
    def __init__(self, **args):
        self.taskSpace = None
        self.args = args
        self.stepName = args.get("StepName", None)
        self.initmodule = inspect.getsourcefile(args.get("Locator", None))
        self.location  = os.path.dirname(self.initmodule)
        self.sandbox = Sandbox(args["TaskName"], self.stepName)

    def sandboxFiles(self):
        """
        _sandboxFiles_

        Get details of sandbox files for this step from the
        WMSandbox

        """
        return self.sandbox.listFiles()


    def getFromSandbox(self, filename, target = None):
        """
        _getFromSandbox_

        Copy a file from the Sandbox area to the Step runtime area.
        Will have same basename as in sandbox unless a target name
        is provided

        """
        sbox = self.sandbox.getFile(filename)
        if target == None:
            target = filename
        destination = os.path.join(self.location, target)
        os.system("/bin/cp %s %s" % (sbox, destination))
        return


    def getWMTask(self):
        """
        _getWMTask_

        Get the WMTask from the TaskSpace

        """
        return self.taskSpace.getWMTask()

    def getWMStep(self):
        """
        _getWMStep_

        Get the WMStep for this Step

        """
        task = self.taskSpace.getWMTask()
        return task.getStep(self.stepName)



#!/usr/bin/env python
"""
_LoadFromID_

Oracle implementation of JobGroup.LoadFromID
"""

__all__ = []



from WMCore.WMBS.MySQL.JobGroup.LoadFromID import LoadFromID as LoadFromIDMySQL

class LoadFromID(LoadFromIDMySQL):
    pass



#!/usr/bin/env python
"""
_DBSBufferUtil_

APIs related to using the DBSBuffer.
"""
from __future__ import print_function

from future.utils import viewitems

import threading
from collections import defaultdict

from WMCore.DAOFactory import DAOFactory
from WMComponent.DBS3Buffer.DBSBufferFile import DBSBufferFile
from WMCore.DataStructs.Run import Run
from WMCore.WMConnectionBase import WMConnectionBase


class DBSBufferUtil(WMConnectionBase):
    """
    APIs related to file addition for DBSBuffer

    """

    def __init__(self):

        myThread = threading.currentThread()
        self.daoFactory = DAOFactory(package="WMComponent.DBS3Buffer",
                                     logger=myThread.logger,
                                     dbinterface=myThread.dbi)

        return

    def loadDBSBufferFilesBulk(self, fileObjs):
        """
        _loadDBSBufferFilesBulk_

        Yes, this is a stupid place to put it.
        No, there's not better place.
        """
        dbsFiles = []

        binds = []
        for f in fileObjs:
            binds.append(f["id"])

        loadFiles = self.daoFactory(classname="DBSBufferFiles.LoadBulkFilesByID")
        results = loadFiles.execute(files=binds, transaction=False)

        for entry in results:
            # Add loaded information
            dbsfile = DBSBufferFile(id=entry['id'])
            dbsfile.update(entry)
            dbsFiles.append(dbsfile)

        for dbsfile in dbsFiles:
            if 'runInfo' in dbsfile:
                # Then we have to replace it with a real run
                for r in dbsfile['runInfo']:
                    run = Run(runNumber=r)
                    run.extend(dbsfile['runInfo'][r])
                    dbsfile.addRun(run)
                del dbsfile['runInfo']
            if 'parentLFNs' in dbsfile:
                # Then we have some parents
                for lfn in dbsfile['parentLFNs']:
                    newFile = DBSBufferFile(lfn=lfn)
                    dbsfile['parents'].add(newFile)
                del dbsfile['parentLFNs']

        return dbsFiles

    def findUploadableDAS(self):
        """
        _findUploadableDAS_

        Find all dataset_algo with uploadable files.
        """
        findDAS = self.daoFactory(classname="FindDASToUpload")
        result = findDAS.execute(transaction=False)

        return result

    def findOpenBlocks(self):
        """
        _findOpenBlocks_

        This should find all blocks.
        """
        openBlocks = self.daoFactory(classname="GetOpenBlocks")
        result = openBlocks.execute(transaction=False)

        return result

    def loadBlocksByDAS(self, das):
        """
        _loadBlocksByDAS_

        Given a DAS, find all the
        blocks associated with it in the
        Open status
        """
        findBlocks = self.daoFactory(classname="LoadBlocksByDAS")
        result = findBlocks.execute(das=das, transaction=False)

        return result

    def loadBlocks(self, blocknames):
        """
        _loadBlocks_

        Given a list of names, load the
        blocks with those names
        """

        if len(blocknames) < 1:
            # Nothing to do
            return []

        findBlocks = self.daoFactory(classname="LoadBlocks")
        result = findBlocks.execute(blocknames, transaction=False)

        return result

    def loadDataset(self, datasetName):
        """
        Given a dataset name, load its information from DBSBuffer
        :param datasetName: string with the dataset name
        :return: a dictionary with the dataset info
        """
        loadDataset = self.daoFactory(classname="ListDataset")
        result = loadDataset.execute(datasetPath=datasetName, transaction=False)
        if not result:
            # this should never happen, but just in case...
            return {}
        return result[0]

    def findUploadableFilesByDAS(self, datasetpath):
        """
        _findUploadableDAS_

        Find all the uploadable files for a given DatasetPath.
        """
        dbsFiles = []

        findFiles = self.daoFactory(classname="LoadDBSFilesByDAS")
        results = findFiles.execute(datasetpath=datasetpath, transaction=False)

        for entry in results:
            # Add loaded information
            dbsfile = DBSBufferFile(id=entry['id'])
            dbsfile.update(entry)
            dbsFiles.append(dbsfile)

        for dbsfile in dbsFiles:
            if 'runInfo' in dbsfile:
                # Then we have to replace it with a real run
                for r in dbsfile['runInfo']:
                    run = Run(runNumber=r)
                    run.extendLumis(dbsfile['runInfo'][r])
                    dbsfile.addRun(run)
                del dbsfile['runInfo']
            if 'parentLFNs' in dbsfile:
                # Then we have some parents
                for lfn in dbsfile['parentLFNs']:
                    newFile = DBSBufferFile(lfn=lfn)
                    dbsfile['parents'].add(newFile)
                del dbsfile['parentLFNs']

        return dbsFiles

    def loadFilesByBlock(self, blockname):
        """
        _loadFilesByBlock_

        Get all files associated with a block
        """
        dbsFiles = []

        findFiles = self.daoFactory(classname="LoadFilesByBlock")
        results = findFiles.execute(blockname=blockname, transaction=False)

        for entry in results:
            # Add loaded information
            dbsfile = DBSBufferFile(id=entry['id'])
            dbsfile.update(entry)
            dbsFiles.append(dbsfile)

        for dbsfile in dbsFiles:
            if 'runInfo' in dbsfile:
                # Then we have to replace it with a real run
                for r in dbsfile['runInfo']:
                    run = Run(runNumber=r)
                    run.extendLumis(dbsfile['runInfo'][r])
                    dbsfile.addRun(run)
                del dbsfile['runInfo']
            if 'parentLFNs' in dbsfile:
                # Then we have some parents
                for lfn in dbsfile['parentLFNs']:
                    newFile = DBSBufferFile(lfn=lfn)
                    dbsfile['parents'].add(newFile)
                del dbsfile['parentLFNs']

        return dbsFiles

    def getCompletedWorkflows(self):
        """
        _getCompletedWorkflows_
        get list of the workflows which are completed.
        completed here is not request manager completed status.
        It indicates all the tasks belongs to this workflow within the agent are completed
        """
        wfCompletedDAO = self.daoFactory(classname="GetCompletedWorkflows")
        result = wfCompletedDAO.execute(transaction=False)

        return result

    def getPhEDExDBSStatusForCompletedWorkflows(self, summary=False):

        """
        _getPhEDxDBSStatuForCompletedWorkflows_
        Check the PhEDEx and DBS upload status for the completed workflow
        """
        wfCompletedStatusDAO = self.daoFactory(classname="CheckStatusForCompletedWorkflows")
        result = wfCompletedStatusDAO.execute(transaction=False)
        if summary:
            result = self.summaryPhEDExDBSStatus(result)

        return result

    def summaryPhEDExDBSStatus(self, data):
        """
        data contains only workflows with completed (in agent not in reqmgr) status
        returns dictionary with kew as workflow and containing dbs/phedex upload status
        """
        summary = defaultdict(dict)
        for workflow, value in viewitems(data):
            # only getting completed workflows
            summary[workflow]["Completed"] = True

            if value["NotInPhEDEx"] == 0 and value["InPhEDEx"] >= 0:
                summary[workflow]["PhEDExInjected"] = True
            else:
                summary[workflow]["PhEDExInjected"] = False

            if value["NotInDBS"] == 0 and value["InDBS"] >= 0:
                summary[workflow]["DBSUploaded"] = True
            else:
                summary[workflow]["DBSUploaded"] = False
        return summary

    def isAllWorkflowCompleted(self):
        """
        check whether all the workflows are completed status in give agent (not globaly in reqmgr)
        This should be used as preconditon for checking draining status
        """
        completeFWFlag = self.daoFactory(classname="IsAllWorkflowsCompleted")
        return completeFWFlag.execute(transaction=False)

    def countOpenBlocks(self):
        """
        check to see if any blocks are open in DBS, reported in drain statistics
        """
        openBlockCount = self.daoFactory(classname="CountOpenBlocks")
        result = openBlockCount.execute()
        return result

    def countFilesByStatus(self, status):
        """
        get counts of files by status, reported in drain statistics
        """
        fileCount = self.daoFactory(classname="CountFilesByStatus")
        result = fileCount.execute(status)
        return result

    def countPhedexNotUploaded(self):
        """
        get counts of files not uploaded to phedex, reported in drain statistics
        """
        phedexCount = self.daoFactory(classname="CountPhedexNotUploaded")
        result = phedexCount.execute()
        return result



#!/usr/bin/env python
"""
_Root_

The root object for a webtools application. It loads all the different views and
starts an appropriately configured CherryPy instance. Views are loaded
dynamically and can be turned on/off via configuration file.

"""
from __future__ import print_function

from builtins import str, bytes
from future.utils import viewitems, listitems

import logging
import os
import socket
import sys
import time
from argparse import ArgumentParser
from pprint import pformat

# CherryPy
import cherrypy
from cherrypy._cplogging import LogManager

# Logging
import WMCore.WMLogging
from Utils.Utilities import lowerCmsHeaders
# configuration and arguments
from WMCore.Agent.Daemon.Create import createDaemon
from WMCore.Agent.Daemon.Details import Details
from WMCore.Agent.Harness import Harness
from WMCore.Configuration import Configuration, ConfigSection
from WMCore.Configuration import loadConfigurationFile
# Factory to load pages dynamically
from WMCore.WMFactory import WMFactory
from WMCore.WebTools.FrontEndAuth import FrontEndAuth, NullAuth
from WMCore.WebTools.Welcome import Welcome

lastTest = ""


def mytime():
    """
    Utility to time stamp start of request handling.
    """
    cherrypy.request.start_time = time.time()


def myproxy(base=None):
    """
    Utility aid in request handling from behind a proxy.
    """
    request = cherrypy.request
    scheme = request.headers.get('X-Forwarded-Proto', request.base[:request.base.find("://")])
    base = request.headers.get('X-Forwarded-Host', base)
    if not base:
        port = cherrypy.request.local.port
        if port == 80:
            base = 'localhost'
        else:
            base = 'localhost:%s' % port

    base = base.split(',')[0].strip()
    if base.find("://") == -1:
        base = scheme + "://" + base
    request.base = base

    xff = request.headers.get('X-Forwarded-For')
    if xff:
        xff = xff.split(',')[0].strip()
        request.remote.ip = xff


class WTLogger(LogManager):
    """
    A logger that logs how we want it to.
    """

    def __init__(self, *args, **kwargs):
        self.host = socket.gethostname()
        LogManager.__init__(self, *args, **kwargs)

    def access(self):
        """
        Record a client access
        """
        request = cherrypy.request
        remote = request.remote
        response = cherrypy.response
        inheaders = lowerCmsHeaders(request.headers)
        outheaders = response.headers
        # identify size of body from HTTP Content-Length header
        rbytes = int(cherrypy.request.headers.get('Content-Length', 0))
        if not rbytes:
            try:
                # request.rfile.rfile.bytes_read is a custom CMS web
                #  cherrypy patch not always available, hence the test
                rbytes = (getattr(request.rfile, 'rfile', None)
                        and getattr(request.rfile.rfile, "bytes_read", None)
                        and request.rfile.rfile.bytes_read) or "-"
            except:
                try:
                    # this will work only when body is read from request
                    rbytes = cherrypy.request.body.fp.bytes_read
                except:
                    rbytes = "-"        
        msg = ('%(t)s %(H)s %(h)s "%(r)s" %(s)s'
               + ' [data: %(i)s in %(b)s out %(T).0f us ]'
               + ' [auth: %(AS)s "%(AU)s" "%(AC)s" ]'
               + ' [ref: "%(f)s" "%(a)s" ]') % {'t': self.time(),
                                                'H': self.host,
                                                'h': remote.name or remote.ip,
                                                'r': request.request_line,
                                                's': response.status,
                                                'i': rbytes,
                                                'b': outheaders.get('Content-Length', '') or "-",
                                                'T': (time.time() - request.start_time) * 1e6,
                                                'AS': inheaders.get("cms-auth-status", "-"),
                                                'AU': inheaders.get("cms-auth-cert",
                                                                    inheaders.get("cms-auth-host", "")),
                                                'AC': getattr(request.cookie.get("cms-auth", None), "value", ""),
                                                'f': inheaders.get("Referer", ""),
                                                'a': inheaders.get("User-Agent", "")}
        self.access_log.log(logging.INFO, msg)


class Root(Harness):
    """
    Create the appropriate cherrypy root object
    """

    def __init__(self, config, webApp=None, testName=""):
        """
        Initialise the object, pull out the necessary pieces of the configuration
        """
        self.homepage = None
        self.mode = 'component'
        self.testName = testName
        if webApp is None:
            Harness.__init__(self, config, compName="Webtools")
            self.appconfig = config.section_(self.config.Webtools.application)
            self.app = self.config.Webtools.application
            self.secconfig = config.component_("SecurityModule")
            self.serverConfig = config.section_("Webtools")
            self.coreDatabase = config.section_("CoreDatabase")
            self.mode = 'standalone'
        else:
            Harness.__init__(self, config, compName=webApp)
            self.appconfig = config.section_(webApp)
            WMCore.WMLogging.setupRotatingHandler(os.path.join(self.appconfig.componentDir, "%s.log" % webApp))
            self.app = webApp
            self.secconfig = getattr(self.appconfig, "security")
            self.serverConfig = config.section_(webApp).section_("Webtools")
            self.coreDatabase = config.section_("CoreDatabase")

        return

    def getLastTest(self):
        global lastTest
        return lastTest

    def setLastTest(self):
        global lastTest
        lastTest = self.testName

    def _validateConfig(self):
        """
        Check that the configuration has the required sections
        """
        config_dict = self.appconfig.dictionary_()
        must_have_keys = ['admin', 'description', 'title']
        for key in must_have_keys:
            msg = "Application configuration '%s' does not contain '%s' key" \
                  % (self.app, key)
            assert key in config_dict, msg

    def _configureCherryPy(self):
        """
        _configureCherryPy_
        Configure the CherryPy server, ignoring items in the configuration file that
        aren't CherryPy configurables.
        """
        configDict = self.serverConfig.dictionary_()

        # If you add configurables update the testLongHandConfigurables in Root_t
        configurables = ['engine', 'hooks', 'log', 'request', 'response',
                         'server', 'tools', 'wsgi', 'checker']
        # Deal with "long hand" configuration variables
        for i in configurables:
            if i in configDict:
                for config_param, param_value in viewitems(configDict[i].dictionary_()):
                    if isinstance(param_value, ConfigSection):
                        # TODO: make this loads better
                        for child_param, child_param_value in viewitems(param_value.dictionary_()):
                            cherrypy.config["%s.%s.%s" % (i, config_param, child_param)] = child_param_value
                    elif isinstance(param_value, (str, bytes, int)):
                        cherrypy.config["%s.%s" % (i, config_param)] = param_value
                    else:
                        raise Exception("Unsupported configuration type: %s" % type(param_value))

        # which we then over write with short hand variables if necessary
        cherrypy.config["server.environment"] = configDict.get("environment", "production")

        # Set up the tools and the logging
        cherrypy.tools.time = cherrypy.Tool('on_start_resource', mytime)
        cherrypy.config.update({'tools.time.on': True})

        if configDict.get("proxy_base", False):
            cherrypy.tools.proxy = cherrypy.Tool('before_request_body', myproxy, priority=30)
            cherrypy.config.update({
                'tools.proxy.on': True,
                'tools.proxy.base': configDict["proxy_base"]
            })

        cherrypy.log = WTLogger()
        cherrypy.config["log.screen"] = bool(configDict.get("log_screen", False))

        cherrypy.config.update({
            'tools.expires.on': True,
            'tools.expires.secs': configDict.get("expires", 300),
            'tools.response_headers.on': True,
            'tools.etags.on': True,
            'tools.etags.autotags': True,
            'tools.encode.on': True,
            'tools.encode.text_only': False,
            'tools.gzip.on': True,
        })

        if cherrypy.config["server.environment"] == "production":
            # If we're production these should be set regardless
            cherrypy.config["request.show_tracebacks"] = False
            cherrypy.config["engine.autoreload.on"] = False
            # In production mode only allow errors at WARNING or greater to the log
            err_lvl = max((configDict.get("error_log_level", logging.WARNING), logging.WARNING))
            acc_lvl = max((configDict.get("access_log_level", logging.INFO), logging.INFO))
            cherrypy.log.error_log.setLevel(err_lvl)
            cherrypy.log.access_log.setLevel(acc_lvl)
        else:
            print()
            print('THIS BETTER NOT BE A PRODUCTION SERVER')
            print()
            cherrypy.config["request.show_tracebacks"] = configDict.get("show_tracebacks", False)
            cherrypy.config["engine.autoreload.on"] = configDict.get("autoreload", False)
            # Allow debug output
            cherrypy.log.error_log.setLevel(configDict.get("error_log_level", logging.DEBUG))
            cherrypy.log.access_log.setLevel(configDict.get("access_log_level", logging.DEBUG))

        default_port = 8080
        if "server.socket_port" in cherrypy.config:
            default_port = cherrypy.config["server.socket_port"]
        cherrypy.config["server.thread_pool"] = configDict.get("thread_pool", 10)
        cherrypy.config["server.accepted_queue_size"] = configDict.get("accepted_queue_size", -1)
        cherrypy.config["server.accepted_queue_timeout"] = configDict.get("accepted_queue_timeout", 10)
        cherrypy.config["server.socket_port"] = configDict.get("port", default_port)
        cherrypy.config["server.socket_host"] = configDict.get("host", "0.0.0.0")
        # A little hacky way to pass the expire second to config
        self.appconfig.default_expires = cherrypy.config["tools.expires.secs"]

        # SecurityModule config
        # Registers secmodv2 into cherrypy.tools so it can be used through
        # decorators
        if self.secconfig.dictionary_().get('dangerously_insecure', False):
            cherrypy.tools.secmodv2 = NullAuth(self.secconfig)
        else:
            cherrypy.tools.secmodv2 = FrontEndAuth(self.secconfig)
            if hasattr(self.secconfig, "default"):
                # If the 'default' section is present, it will force the
                # authn/z to be called even for non-decorated methods
                cherrypy.config.update({'tools.secmodv2.on': True,
                                        'tools.secmodv2.role': self.secconfig.default.role,
                                        'tools.secmodv2.group': self.secconfig.default.group,
                                        'tools.secmodv2.site': self.secconfig.default.site})
        cherrypy.config.update({'tools.cpstats.on': configDict.get('cpstats', False)})
        cherrypy.config.update({'server.statistics': configDict.get('cpstats', False)})
        cherrypy.log.error_log.debug('Application %s initialised in %s mode', self.app, self.mode)
        cherrypy.log.access_log.info("Final CherryPy configuration: %s" % pformat(cherrypy.config))

    def _loadPages(self):
        """
        Load up all the pages in the configuration
        """
        factory = WMFactory('webtools_factory')

        globalconf = self.appconfig.dictionary_()
        del globalconf['views']
        the_index = ''
        if 'index' in globalconf:
            the_index = globalconf['index']
            del globalconf['index']

        for view in self.appconfig.views.active:
            # Iterate through each view's configuration and instantiate the class
            if view._internal_name != the_index:
                if 'instances' in globalconf:
                    for instance in globalconf['instances']:
                        self._mountPage(view, globalconf, factory, instance)
                else:
                    self._mountPage(view, globalconf, factory)

        if hasattr(self.appconfig.views, 'maintenance'):
            # for i in self.appconfig.views.maintenance:
            # TODO: Show a maintenance page with a 503 Service Unavailable header
            pass

    def _generate_config(self, view, globalconf, instance=None, is_index=False):
        """
        _generate_config_
        Create the configuration for a page by combining it's configuration
        with the global one
        """
        config = Configuration()
        view_config = config.component_(view._internal_name)
        view_config.application = self.app

        view_dict = view.dictionary_()
        for k in globalconf:
            # Add the global config to the view
            view_config.__setattr__(k, globalconf[k])
        for k in view_dict:
            # overwrite global if the local config is different
            view_config.__setattr__(k, view_dict[k])

        # TODO: remove bits we don't need

        if instance:
            # record the instance into the view's configuration
            view_config.instance = instance
            if hasattr(view, 'database') and hasattr(view.database, 'instances'):
                db_cfg = view.database.section_('instances')
                view_config.section_('database')
                view_config.database = db_cfg.section_(instance)

            if hasattr(view, 'security') and hasattr(view.security, 'instances'):
                security_cfg = view.security.section_('instances')
                view_config.section_('security')
                view_config.security = security_cfg.section_(instance)

        if 'database' in view_config.dictionary_():
            if not isinstance(view_config.database, (str, bytes)):
                if len(view_config.database.listSections_()) == 0:
                    if len(self.coreDatabase.listSections_()) > 0:
                        view_config.database.connectUrl = self.coreDatabase.connectUrl
                        if hasattr(self.coreDatabase, "socket"):
                            view_config.database.socket = self.coreDatabase.socket
        return view_config

    def _mountPage(self, view, globalconf, factory, instance=None, is_index=False):
        """
        _mountPage_
        Add the page to the CherryPy tree.
        """
        if is_index:
            mount_point = os.path.join('/', self.app.lower())
        elif instance:
            mount_point = os.path.join('/', self.app.lower(), instance, view._internal_name)
        else:
            mount_point = os.path.join('/', self.app.lower(), view._internal_name)
        view_config = self._generate_config(view, globalconf, instance, is_index)
        # component now contains the full configuration (global + view)
        # use this throughout

        cherrypy.log.error_log.debug("Loading %s", view_config._internal_name)
        # Load the object
        obj = factory.loadObject(view_config.object, view_config, getFromCache=False)
        # Attach the object to cherrypy's tree, at the name of the component
        cherrypy.tree.mount(obj, mount_point)
        msg = "%s available on %s/%s" % (view_config._internal_name,
                                         cherrypy.server.base(),
                                         mount_point)
        cherrypy.log.error_log.info(msg)

    def _makeIndex(self):
        """
        Create an index page, either from the configured page or a generic default
        welcome page.
        """
        globalconf = self.appconfig.dictionary_()
        if hasattr(self.appconfig, 'index'):
            factory = WMFactory('webtools_factory')
            view = getattr(self.appconfig.views.active, globalconf['index'])
            del globalconf['views']
            del globalconf['index']
            self._mountPage(view, globalconf, factory, is_index=True)

        else:
            cherrypy.log.error_log.info("No index defined for %s - instantiating default Welcome page", self.app)
            namesAndDocstrings = []
            # make a default Welcome
            for view in self.appconfig.views.active:
                if not getattr(view, "hidden", False):
                    viewName = view._internal_name
                    if 'instances' in globalconf:
                        for instance in globalconf['instances']:
                            mount_point = '/%s/%s/%s' % (self.app.lower(), instance, viewName)
                            viewObj = cherrypy.tree.apps[mount_point].root
                            docstring = viewObj.__doc__
                            namesAndDocstrings.append(('%s/%s' % (instance, viewName), docstring))
                    else:
                        mount_point = '/%s/%s' % (self.app.lower(), viewName)
                        viewObj = cherrypy.tree.apps[mount_point].root
                        docstring = viewObj.__doc__
                        namesAndDocstrings.append((viewName, docstring))
            cherrypy.tree.mount(Welcome(namesAndDocstrings), "/%s" % self.app.lower())

    def start(self, blocking=True):
        """
        Configure and start the server
        """
        self._validateConfig()
        self._configureCherryPy()
        self._loadPages()
        self._makeIndex()
        cherrypy.server.httpserver = None
        cherrypy.engine.start()
        if blocking:
            cherrypy.engine.block()

    def startComponent(self):
        """
        _startComponent_

        Called by the WMAgent harness code.  This will never return.
        """
        self.start()

    def stop(self):
        """
        Stop the server
        """
        cherrypy.engine.exit()
        cherrypy.engine.stop()

        # Ensure the next server that's started gets fresh objects
        for name, server in listitems(getattr(cherrypy, 'servers', {})):
            server.unsubscribe()
            del cherrypy.servers[name]


if __name__ == "__main__":
    parser = ArgumentParser()
    parser.add_argument("-i", "--ini", dest="inifile", default=False,
                        help="write the configuration to FILE", metavar="FILE")
    parser.add_argument("-v", "--verbose",
                        action="store_true", dest="verbose", default=False,
                        help="Be more verbose")
    parser.add_argument("-d", "--daemonise",
                        action="store_true", dest="daemon", default=False,
                        help="Daemonise the cherrypy process, and return the PID")
    parser.add_argument("-s", "--status",
                        action="store_true", dest="status", default=False,
                        help="Return the status of the daemon")
    parser.add_argument("-k", "--kill",
                        action="store_true", dest="kill", default=False,
                        help="Kill the daemon")
    parser.add_argument("-t", "--terminate",
                        action="store_true", dest="terminate", default=False,
                        help="Terminate the daemon (kill, wait, kill -9)")
    opts = parser.parse_args()

    if not opts.inifile:
        sys.exit('No configuration specified')
    cfg = loadConfigurationFile(opts.inifile)

    component = cfg.Webtools.application
    workdir = getattr(cfg.Webtools, 'componentDir', '/tmp/webtools')
    if workdir is None:
        workdir = '/tmp/webtools'
    root = Root(cfg)
    if opts.status:
        daemon = Details('%s/Daemon.xml' % workdir)

        if not daemon.isAlive():
            print("Component:%s Not Running" % component)
        else:
            print("Component:%s Running:%s" % (component, daemon['ProcessID']))
    elif opts.kill:
        daemon = Details('%s/Daemon.xml' % workdir)
        daemon.kill()
        daemon.removeAndBackupDaemonFile()
    elif opts.terminate:
        daemon = Details('%s/Daemon.xml' % workdir)
        daemon.killWithPrejudice()
        daemon.removeAndBackupDaemonFile()
    elif opts.daemon:
        createDaemon(workdir)
        root.start(False)
    else:
        root.start()



#!/usr/bin/env python
"""
_GetNumberOfJobsPerSite_

Oracle implementation of Jobs.GetNumberOfJobsPerSite
"""

__all__ = []



import logging

from WMCore.WMBS.MySQL.Jobs.GetNumberOfJobsPerSite import GetNumberOfJobsPerSite as MySQLGetNumberOfJobsPerSite

class GetNumberOfJobsPerSite(MySQLGetNumberOfJobsPerSite):
    """
    Identical to MySQL version

    """
